# Copyrite IBM 2022, 2025
# IBM Confidential
# Assisted by watsonx Code Assistant

import base64, datetime, multiprocessing, os, sys, random, threading, time, secrets, typing, logging, math
import cbor2 as cbor
from enum import Enum, IntEnum
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.fernet import Fernet

try:
    from soft_fido2.message_queues import QueueMessageType, MessageQueue
    from soft_fido2.uhid_device import UserDevice, BaseStructure, bcolors, dump_bytes, colour_print
    from soft_fido2.key_pair import KeyPair, KeyUtils
    from soft_fido2.authenticator import Fido2Authenticator
    from soft_fido2.cert_utils import CertUtils
    from soft_fido2.symmetric_key import SymmetricKey
except ImportError:
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))
    from message_queues import QueueMessageType, MessageQueue
    from uhid_device import UserDevice, BaseStructure, bcolors, dump_bytes, colour_print
    from key_pair import KeyPair, KeyUtils
    from authenticator import Fido2Authenticator
    from cert_utils import CertUtils
    from symmetric_key import SymmetricKey

#max usb data frame size
MAX_DATA_FRAME = 64

class AuthenticatorAPI(object):
    '''
    AuthenticatorAPI wraps the Fido2Authenticator with management for a persistent key/ca certificate stored in
    an encrypted file.

    The default location for this is in the ~/.fido2 directory for the user running this module, however this
    can be overwritten by environment properties.

    All keys generated by this authenticator are created so that the credential id is the encrypted key
    used to generate the attestation. This is done to allow this Authenticator to restore any key created with
    its ca key / AES key combination without the need to persist every individual private key used to 
    generate a new credential.
    '''

    _exp_time = 30

    _open_keys = {}

    _watchdog = None
    _lock = multiprocessing.Lock()

    _pin_token_kp: typing.Optional[KeyPair] = None
    _pin_retry = 5

    _quit = False

    def __new__(cls):
        cls._watchdog = threading.Thread(target=cls._token_expiry_check)
        cls._watchdog.start()
        cls._lock.acquire()
        if cls._pin_token_kp == None:
            cls._pin_token_kp = KeyPair.generate_ecdsa()#P-256 key, cose key type -25
        cls._lock.release()

    @classmethod
    def _token_expiry_check(cls):
        '''
        Expire in-memory passkeys handled by open CIDs after a short duration since use for authentication as they would no longer be used
        '''
        while not cls._quit:
            time.sleep(0.005)
            cls._lock.acquire()
            cid_list = list(cls._open_keys.keys())
            for cid in cid_list:
                if math.floor(time.time() - cls._open_keys[cid]["tStart"]) == cls._exp_time:
                    cls._open_keys.pop(cid)
                    colour_print(colour=bcolors.FAIL, component='Authenticator_token_expiry_check',
                                 msg='CID {} has expired!\nExisting tokens: {}'.format(cid, cls._open_keys))
            cls._lock.release()


    @classmethod
    def has_cached_up(cls, cid) -> bool:
        cls._lock.acquire()
        try:
            if cid in cls._open_keys:
                return bool(cls._open_keys[cid].get("up", False))
            return False
        finally:
            cls._lock.release()

    @classmethod
    def cache_up(cls, cid, up: bool):
        if cid in cls._open_keys:
            cls._lock.acquire()
            cls._open_keys[cid]["up"] = up
            cls._lock.release()

    @classmethod
    def get_pin_auth_token(cls, cid):
        cls._lock.acquire()
        try:
            # Find CID in open passkeys and then attempt to acquire pinUvAuthToken
            open_key = AuthenticatorAPI._open_keys.get(cid, {})
            return open_key.get('pinAuth')
        finally:
            cls._lock.release()

    @classmethod
    def _validate_pin(cls, pinHash: bytes, cid: bytes) -> typing.Optional[bytes]:
        """
        Validates a PIN by attempting to decrypt passkey files in the FIDO_HOME directory.
        
        If a valid passkey file is found, it loads the certificate and key pair,
        stores the information in the class's _open_keys dictionary with the channe
        id as the key, and returns a generated PIN authentication token.
        
        Args:
            pinHash: The hash of the PIN to validate
            cid: The channel ID to associate with the opened keys
            
        Returns:
            A PIN authentication token if validation succeeds, None otherwise
        """
        # Check if FIDO_HOME environment variable exists and directory is accessible
        if not cls._is_fido_home_valid():
            return None
        fido_home_dir = os.path.realpath(os.environ["FIDO_HOME"])
        for passkey_file in cls._get_passkey_files(fido_home_dir):
            try:
                # Attempt to load and validate the passkey
                auth_token = cls._process_passkey_file(passkey_file, pinHash, cid)
                if auth_token:
                    return auth_token
            except Exception as e:
                colour_print(
                    colour=bcolors.WARNING, 
                    component='Authenticator_validate_pin',
                    msg=f'Failed to process {os.path.basename(passkey_file)}:\n{e}'
                )
                continue
        
        # No valid passkey file found
        colour_print(
            colour=bcolors.FAIL, 
            component='Authenticator_validate_pin',
            msg='No valid pin found!'
        )
        return None

    @classmethod
    def _is_fido_home_valid(cls) -> bool:
        """
        Checks if the FIDO_HOME environment variable exists and points to a valid directory.
        
        Returns:
            bool: True if FIDO_HOME is valid, False otherwise
        """
        if "FIDO_HOME" not in os.environ:
            logging.debug("FIDO_HOME not set, can't do much . . .")
            return False
            
        fido_home_dir = os.path.realpath(os.environ["FIDO_HOME"])
        if not os.path.exists(fido_home_dir):
            logging.debug("FIDO_HOME directory not found, can't do much . . .")
            return False
            
        return True

    @classmethod
    def _get_passkey_files(cls, directory: str) -> typing.List[str]:
        """
        Returns a list of .passkey files in the specified directory.
        
        Args:
            directory: The directory to search for .passkey files
            
        Returns:
            A list of full paths to .passkey files
        """
        passkey_files = []
        for filename in os.listdir(directory):
            if filename.endswith('.passkey'):
                passkey_files.append(os.path.join(directory, filename))
            else:
                colour_print(
                    colour=bcolors.WARNING, 
                    component='Authenticator_validate_pin',
                    msg=f'{filename} has invalid file type'
                )
        return passkey_files

    @classmethod
    def _process_passkey_file(cls, passkey_file: str, pinHash: bytes, cid: bytes) -> typing.Optional[bytes]:
        """
        Attempts to decrypt and process a passkey file.
        
        Args:
            passkey_file: Path to the passkey file
            pinHash: The hash of the PIN to validate
            cid: The channel ID to associate with the opened keys
            
        Returns:
            A PIN authentication token if processing succeeds, None otherwise
            
        Raises:
            Various exceptions if file processing fails
        """
        # Load and decrypt the passkey file
        passkey = KeyUtils._load_passkey(pinHash, passkey_file)
        
        colour_print(
            colour=bcolors.OKPINK, 
            component='Authenticator_validate_pin',
            msg='Pin decrypted a .passkey file'
        )
        
        # Extract certificate and key pair
        ca_x5c = passkey.get('x5c')
        key = passkey.get('key')
        if not isinstance(key, ec.EllipticCurvePrivateKey):
            raise ValueError("Key must be an ECPrivateKey")
        key_pair = KeyPair(key, key.public_key())
        #seed = passkey.get('seed')
        
        # Reset PIN retry counter
        cls._pin_retry = 5
        
        # Generate authentication token
        pin_auth_token = secrets.token_bytes(32)
        
        # Store the opened keys
        cls._lock.acquire()
        try:
            cls._open_keys[cid] = {
                'x5c': ca_x5c,
                'kp': key_pair,
                'file': passkey_file,
                'ph': pinHash,
                'pinAuth': pin_auth_token,
                'tStart': time.time() # Get Unix timestamp to compare against for watchdog thread (expiry timer)
            }
            return pin_auth_token
        finally:
            cls._lock.release()


    @classmethod
    def get_pin_cose_key(cls, pin_req, cid):
        if cls._pin_token_kp == None:
            cls._pin_token_kp = KeyPair.generate_ecdsa()
        return {1: KeyUtils.get_cose_key(cls._pin_token_kp.get_public(), hashes.SHA256(), eckx=True)}

    @classmethod
    def get_pin_retries(cls, pin_req, cid):
        cls._pin_retry -= 1
        return {3: cls._pin_retry}

    @classmethod
    def decapsulate(cls, ecCoseKey):
        cose_type_to_curve_map = { #These are kind of made up, as per 
                    #https://fidoalliance.org/specs/fido-v2.1-ps-20210615/fido-client-to-authenticator-protocol-v2.1-ps-errata-20220621.html#pinProto1
                    -25: ec.SECP256R1,
                    -26: ec.SECP521R1
                }
        ec_pub_numbs = ec.EllipticCurvePublicNumbers(KeyUtils._bytes_to_long(ecCoseKey[-2]),
                            KeyUtils._bytes_to_long(ecCoseKey[-3]), 
                            cose_type_to_curve_map[ecCoseKey[3]]())
        pubkey = ec_pub_numbs.public_key()
        shared_point = cls._pin_token_kp.get_private().exchange(ec.ECDH(), pubkey) # type: ignore don't decapsulate without a key exchange
        hasher = hashes.Hash(hashes.SHA256())
        hasher.update(shared_point)
        return hasher.finalize()

    @classmethod
    def get_pin_token(cls, pin_req, cid):
        #https://fidoalliance.org/specs/fido-v2.1-ps-20210615/fido-client-to-authenticator-protocol-v2.1-ps-errata-20220621.html#getPinToken
        logging.debug(f"pin_req: {pin_req}")
        platform_cose_key = pin_req[3]
        pin_hash_enc = pin_req[6]
        colour_print(colour=bcolors.OKPINK, component='Authenticator.get_pin_token', 
                     msg='plat cose key: {}; pinHashEnc: {}'.format(platform_cose_key, pin_hash_enc))
        sharedSecret = cls.decapsulate(platform_cose_key)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.get_pin_token',
                     msg='shared secret: {};'.format(sharedSecret))
        cipher = Cipher(algorithms.AES256(sharedSecret), modes.CBC(bytes([0] * 16))) # nosemgrep part of the CTAP2 spec
        decryptor = cipher.decryptor() # nosemgrep
        pin_hash = decryptor.update(pin_hash_enc) + decryptor.finalize()
        pinAuthToken = cls._validate_pin(pin_hash, cid)
        if pinAuthToken != None:
            encryptor = cipher.encryptor()
            pinAuthTokenEnc = encryptor.update(pinAuthToken) + encryptor.finalize()
            return {2: pinAuthTokenEnc}
        return None


    @classmethod
    def attestation_out(cls, clientDataHash, rp, user, pkCredsParams, excludeList, exts, cid):
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='open keys: {}'.format(cls._open_keys))
        if not cid in cls._open_keys.keys():
            return CBORCommand.CBORStatusCode.CTAP2_ERR_OPERATION_DENIED, None, None
        passkey: dict = cls._open_keys[cid]
        ca_kp = passkey.get('kp')
        if not isinstance(ca_kp, KeyPair): # !panic
            colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', msg="panic!")
            return CBORCommand.CBORStatusCode.CTAP1_ERR_OTHER, None, None
        seed = KeyUtils.get_passkey_seed(rp['id'].encode(), ca_kp.get_private())
        #logging.debug(f"seed : {seed}")
        fkey = Fernet(seed)
        kp = KeyPair.generate_ecdsa()
        resCreds = KeyUtils._load_passkey(passkey['ph'], passkey['file']).get('res.creds') #
        #logging.debug(f"checking for existing registration for {rp} and {user} in {resCreds}")
        if resCreds: # Check for existing rpID:userID
            for cred in resCreds:
                if rp['id'] == cred['rp.id'] and user['id'] == cred['user.id']:
                    colour_print(colour=bcolors.FAIL, component='Authenticator.attestation_out',
                                 msg='existing rpID and userID found: {}, {}'.format(rp['id'], user['id']))
                    return CBORCommand.CBORStatusCode.CTAP2_ERR_CREDENTIAL_EXCLUDED, None, None
                #else:
                #    logging.debug(f"{rp['id']} != {cred['rp.id']} or {user['id']} != {cred['user.id']}")
        authenticator = Fido2Authenticator(keyPair=kp, caKeyPair=ca_kp, 
                                            caCert=passkey.get('x5c'), fKey=fkey)
        credId = authenticator._get_credential_id_bytes(kp)
        pk = {'rp': rp} # authenticator.py expects this structure
        authData = authenticator.build_authenticator_data(pk, 'packed', kp, True, up=True, be=False, bs=False)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='credId: {}; toSign: {}'.format(credId, 
                            base64.b64encode(bytes([*authData, *clientDataHash ])).decode()))
        attStmt = authenticator.build_packed_attestation_statement('packed', 
                                                    clientDataHash, authData, None, kp,)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='attStmt: {}'.format(attStmt))
        #Since we set UV we can make these resident keys
        KeyUtils.update_passkey({'cred.id': credId, 'user.id': user['id'], 'rp.id': rp['id']},
                                passkey['ph'], passkey['file'])
        return None, authData, attStmt


    @classmethod
    def _maybe_next_assertion(cls, rpId, ca_kp, ca_x5c, clientDataHash, cred): 
        seed = KeyUtils.get_passkey_seed(rpId.encode(), ca_kp.get_private())
        #logging.debug(f"seed : {seed}")
        fkey = Fernet(seed)
        #skey = SymmetricKey(seed.decode())
        #Create the authenticator from the raw id, creating the key pair will fail if we don't own the credId
        b64CredId = base64.urlsafe_b64encode(cred.get('id'))
        decryptedKp = Fido2Authenticator._get_key_pair_from_credential_id(b64CredId, fkey)
        colour_print(colour=bcolors.OKPINK, component='FIDO2Authenticator.assertion_outputs',
                        msg='We have a usable key, sign the challenge')
        _authenticator = Fido2Authenticator(keyPair=decryptedKp, credId=b64CredId, aaguid=[0] * 16, 
                                            caKeyPair=ca_kp, caCert=ca_x5c, fKey=fkey)
        credential = {
                "id": _authenticator._get_credential_id_bytes(_authenticator.kp), #Should be set by __init__()
                "type" : "public-key"
            }
        #Generate the assertion response data
        authData = _authenticator.build_authenticator_data({'rpId': rpId}, 'packed', _authenticator.kp,
                                                                 True, up=True, be=False, bs=False)
        #Sign it, authenticator method b64 encodeds the ba so we decode it right away
        sig = _authenticator.assertion_signature(authData, clientDataHash, _authenticator.kp)
        userHandle = cred.get("user")
        return None, credential, authData, sig, userHandle


    @classmethod
    def assertion_out(cls, rpId, clientDataHash, allowedList, exts, cid):
        if not cid in cls._open_keys.keys():
            return CBORCommand.CBORStatusCode.CTAP2_ERR_OPERATION_DENIED, None, None, None, None
        passkey: dict = cls._open_keys[cid]
        ca_x5c = passkey.get('x5c')
        ca_kp = passkey.get('kp')
        if not isinstance(ca_kp, KeyPair): # !panic
            return CBORCommand.CBORStatusCode.CTAP1_ERR_OTHER, None, None, None, None
        resCreds = KeyUtils._load_passkey(passkey['ph'], passkey['file']).get('res.creds')
        if resCreds != None:
            colour_print(colour=bcolors.OKPINK, component='FIDO2Authenticator.assertion_outputs',
                         msg='passkey has resident credentials, adding them to allowed list')
            for cred in resCreds:
                if cred.get('rp.id') == rpId:
                    allowedList += [{'id': cred.get('cred.id'), 
                                     'user': cred.get('user.id')}]
        for cred in allowedList:
            try:
                #logging.debug(f"Try {cred}")
                return cls._maybe_next_assertion(rpId, ca_kp, ca_x5c, clientDataHash, cred)
            except Exception as e:
                colour_print(colour=bcolors.FAIL, component='FIDO2Authenticator.assertion_outputs',
                             msg='Could not retrieve key pair from credential id {}'.format(cred))
                logging.exception(e, stack_info=True)
                continue
        return CBORCommand.CBORStatusCode.CTAP2_ERR_NO_CREDENTIALS, None, None, None, None

    @classmethod
    def quit(cls):
        cls._quit = True
        if cls._watchdog:
            cls._watchdog.join()

class CBORCommand(object):

    class CommandByte(Enum):
        MAKE_CREDENTIAL = 0x1
        GET_NEXT_ASSERTION = 0x2
        GET_INFO = 0x4
        CLIENT_PIN = 0x6
        RESET = 0x7
        CREDENTIAL_MANAGEMENT = 0x9
        AUTHENTICATOR_SELECTION = 0xB
        AUTHENTICATOR_CONFIG = 0xD

        def __repr__(self):
            return str(self.value)

    class CBORStatusCode(IntEnum):
        CTAP2_OK = 0x0
        CTAP1_ERR_INVALID_COMMAND = 0x01
        CTAP1_ERR_TIMEOUT = 0x05
        CTAP2_ERR_INVALID_CBOR = 0x12
        CTAP2_ERR_MISSING_PARAMETER = 0x14
        CTAP2_ERR_CREDENTIAL_EXCLUDED = 0x19
        CTAP2_ERR_OPERATION_DENIED = 0x27
        CTAP2_ERR_NO_CREDENTIALS = 0x2E
        CTAP2_ERR_PIN_INVALID = 0x31
        CTAP2_ERR_PIN_AUTH_INVALID = 0x33
        CTAP2_ERR_PUAT_REQUIRED = 0x36
        CTAP1_ERR_OTHER = 0x7F


    cid = 0xFFFFFFFF
    request = []
    response: list = []
    response_segment = 0
    response_ready = False
    length = 0
    request_segment = 0
    cmd = None
    ctaphid_cmd = 0
    bcnt = 0
    _pending = None

    def __init__(self, cid, ba, skip_init=False):
        self.cid = cid
        if ba == None and skip_init == True:
            return #Create an empty command as we will directly set the response buffer later with the assigned CID.
        if len(ba) <= 1:
            colour_print(colour=bcolors.OKYELLOW, component='CBORCommand.__init__', 
                    msg="Byte Array must be at least one byte long")
        # Length of the incoming CBOR message (total).
        self.length = int.from_bytes(ba[0:2], 'big') - 1 # subtract CMD byte
        # Request buffer. This stores the incoming CBOR message and grows until all segments have been received
        self.request = ba[3:]
        # Track then number of response segments transmitted, the number transmitted in the continue sequence packet
        # should always be one less than this number
        self.response_segment = 0
        # Track the number of request segments received
        self.request_segment = 0
        # Response buffer. This stores the outgoing response to the received CBOR message and shrinks until the entire
        # response has been transmitted
        self.response = []
        # Command received in CTAPHID frame, this is likely 0x90 (CBOR_MSG) but might be different
        self.ctaphid_cmd = 0
        # Authenticator API command byte received in initial packet
        self.cmd = self.CommandByte(int.from_bytes(ba[2:3], 'big'))
        #Length of the payload bytes
        self.bcnt = 0
        # Signal that the response buffer is ready to be sent back to the client
        self.response_ready = False
        colour_print(colour=bcolors.OKPURPLE, component='CBORCommand.__init__', 
                msg="command {}; length {}; self.request[{}]".format(self.cmd, self.length, len(self.request)))
        if self.length >= len(self.request):
            colour_print(colour=bcolors.OKPURPLE, component='CBORCommand.__init__', 
                    msg="request is segmented, wait for the whole message")
        else: #We have the whole message
            self.unpack()
            dump_bytes(self.response, colour=bcolors.OKPINK,
                       component='CBORCommand.__init__', msg='CTAP response')

    def append_segment(self, seg_buf):
        self.request += seg_buf
        if len(self.request) >= self.length:
            self.unpack()
            dump_bytes(self.response, colour=bcolors.OKPINK,
                       component='CBORCommand.append_segment', msg='CTAP response')
        else:
            self.request_segment += 1

    def _error(self, ba):
        self.response = list(self.CBORStatusCode.CTAP1_ERR_INVALID_COMMAND.to_bytes(1, 'big'))
        self.bcnt = 0
        self.response_ready = True

    #Return CBOR response if entire command has been received or None if still 
    #waiting for segments
    def unpack(self):
        if self.cmd == None:
            return self._error(None)
        return {
            self.CommandByte.MAKE_CREDENTIAL: self._make_cred,
            self.CommandByte.GET_NEXT_ASSERTION: self._get_assertion,
            self.CommandByte.GET_INFO: self._get_info,
            self.CommandByte.CLIENT_PIN: self._client_pin
            }.get(self.cmd, self._error)(bytes(self.request))

    def _set_rsp_fields(self, rsp=[]):
        self.response = rsp
        self.bcnt = len(rsp)
        self.response_ready = True

    def get_rsp_seg(self, num_bytes):
        if not isinstance(num_bytes, int):
            raise RuntimeError("panic!")
        self.response_segment += 1
        #sequence is offset by two to account for init pkt and zero index start for continue sequence
        seg_num = max(self.response_segment - 2, 0)
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', 
                msg='self.response_segment = {}'.format(self.response_segment))
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', 
                msg='self.response_segment - 2 = {}'.format(self.response_segment - 2))
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', 
                msg='segment number = {}'.format(seg_num))
        seg = self.response
        if num_bytes >= len(self.response):
            self.response = []
        else:
            seg = self.response[:num_bytes]
            self.response = self.response[num_bytes:]
        return seg, seg_num

    @classmethod
    def set_pending(cls, pending):
        cls._pending = pending

    def gather_user_presence(self):
        if os.environ.get('SOFT_FIDO2_SKIP_UP', 'False').lower() in ['y', 'yes', '1', 'true', 't']:
            colour_print(colour=bcolors.WARNING, component='Authenticator.gather_user_presence', 
                    msg='Skipping user presence check')
            return True
        elif AuthenticatorAPI.has_cached_up(self.cid):
            return True

        start_time = time.time()
        MessageQueue.notify_auth.queue.clear()
        MessageQueue.notify_sysapp.put(QueueMessageType.USER_REQUEST)
        msg = None
        worker = KeepAliveWorker(self._pending, self.cid)
        worker.start()
        current_time = time.time()
        while not msg and current_time - start_time < 60:
            time.sleep(0.002)
            current_time = time.time()
            if MessageQueue.notify_auth.qsize() > 0:
                msg = MessageQueue.notify_auth.get()
        worker.interrupt()
        worker.join()
        if msg == QueueMessageType.USER_RESPONSE_ACCEPT:
            AuthenticatorAPI.cache_up(self.cid, True)
            return True
        return False

    def _verify_pin_token(self, clientDataHash, pinUvAuthParam):
        if pinUvAuthParam not in [None, b'']:
            pinAuth = AuthenticatorAPI.get_pin_auth_token(self.cid)
            # Verify token using client data hash
            h = hmac.HMAC(pinAuth, hashes.SHA256())
            h.update(clientDataHash)
            sig = h.finalize()
            if pinUvAuthParam != sig[:16]: # token not valid if the first 16 bytes of signature doesn't match pinUvAuthParam
                return False
            return True
        else:
            return False

    def _client_pin(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorClientPIN
        pin_sub_cmds = { 
                      1: AuthenticatorAPI.get_pin_retries,
                      2: AuthenticatorAPI.get_pin_cose_key,
            #SET_PIN = 0x3
            #CHANGE_PIN = 0x4
                      5: AuthenticatorAPI.get_pin_token
                }
        req_data = cbor.loads(ba)
        colour_print(colour=bcolors.OKGREEN, component='CBORCommand._client_pin',
                     msg='Packet request: {}'.format(req_data))
        sub_cmd = req_data[2]
        colour_print(colour=bcolors.OKGREEN, component='CBORCommand._client_pin',
                     msg='pin sub_cmd: {}'.format(sub_cmd))
        rsp = pin_sub_cmds[sub_cmd](req_data, self.cid)
        result = (self.CBORStatusCode.CTAP2_ERR_PIN_INVALID).to_bytes(1, 'big')
        if rsp != None:
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes(1, 'big') + cbor.dumps(rsp)
        return self._set_rsp_fields(list(result))

    # authenticatorGetInfo takes no inputs so return immediately
    def _get_info(self, ba):
        result = {
            0x01: ["U2F_V2", "FIDO_2_0"],
            0x02: ['hmac-secret'],
            #0x03: b"\x13\x37\xF1\xD0" * 4,
            0x03: b"\x00" * 16,
            0x04: {'rk': True, 'up': True, 'plat': False, 'clientPin': True},
            0x05: 1200,
            0x06: [1]
        }
        result = bytes( (self.CBORStatusCode.CTAP2_OK).to_bytes(1, 'big') + cbor.dumps(result) )
        logging.debug(f"len: {len(result)}")
        return self._set_rsp_fields(list(result))

    def _make_cred(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorMakeCredential
        if self.gather_user_presence() == False:
            return self._set_rsp_fields( list((self.CBORStatusCode.CTAP2_ERR_OPERATION_DENIED).to_bytes(1, 'big')) )
        req = cbor.loads(ba)
        colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                     msg='CBOR request {}'.format(req))
        for prop in [(0x01, 'clientDataHash'), (0x02, 'rp'), (0x03, 'user'), (0x04, 'pubkeyCredParams'), (0x08, 'pinAuth')]:
            if not prop[0] in req.keys():
                colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                             msg='{} missing from request:\n{}'.format(prop[1], cbor.dumps(req)))
                logging.debug("Missing required property %s" % prop[1])
                return self._set_rsp_fields( list((self.CBORStatusCode.CTAP2_ERR_MISSING_PARAMETER).to_bytes(1, 'big')) )
        result = (self.CBORStatusCode.CTAP2_ERR_PUAT_REQUIRED).to_bytes(1, 'big')
        # Request and validate pin-auth
        if not self._verify_pin_token(req.get(0x01), req.get(0x08)):
            if self.cid in AuthenticatorAPI._open_keys:
                result = (self.CBORStatusCode.CTAP2_ERR_PIN_AUTH_INVALID).to_bytes(1, 'big')
            return self._set_rsp_fields(list(result))
        error, authData, attStmt = AuthenticatorAPI.attestation_out(req.get(0x01), req.get(0x02), req.get(0x03),
                                            req.get(0x04), req.get(0x05), req.get(0x06), self.cid)
        result = (self.CBORStatusCode.CTAP1_ERR_OTHER).to_bytes(1, 'big')
        if error:
            result = error.to_bytes(1, 'big')
        if authData and attStmt:
            rsp = {
                0x01: 'packed', #fmt
                0x02: authData,
                0x03: attStmt
            }
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes(1, 'big') + cbor.dumps(rsp)
        return self._set_rsp_fields(list(result))


    def _get_assertion(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorGetAssertion
        if self.gather_user_presence() == False:
            return self._set_rsp_fields( list((self.CBORStatusCode.CTAP2_ERR_OPERATION_DENIED).to_bytes(1, 'big')) )
        req = cbor.loads(ba)
        colour_print(colour=bcolors.FAIL, component='CBORCommand._get_assertion',
                     msg='CBOR request {}'.format(req))
        for prop in [(0x01, 'rpId'), (0x02, 'clientDataHash'), (0x06, 'pinAuth')]:
            if not prop[0] in req:
                colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                             msg='{} missing from request:\n{}'.format(prop[1], cbor.dumps(req)))
                logging.debug("Missing required property %s" % prop[1])
                return self._set_rsp_fields( list((self.CBORStatusCode.CTAP2_ERR_MISSING_PARAMETER).to_bytes(1, 'big')) )
        result = (self.CBORStatusCode.CTAP1_ERR_OTHER).to_bytes(1, 'big')
        # Request and validate pin-auth
        if not self._verify_pin_token(req.get(0x02), req.get(0x06)):
            if self.cid in AuthenticatorAPI._open_keys:
                result = (self.CBORStatusCode.CTAP2_ERR_PIN_AUTH_INVALID).to_bytes(1, 'big')
            return self._set_rsp_fields(list(result))
        error, credential, authData, signature, userHandle = AuthenticatorAPI.assertion_out(req.get(0x01), 
                                                req.get(0x02), req.get(0x03, []), req.get(0x04, {}), self.cid)
        if error:
            result = error.to_bytes(1, 'big')
        elif credential and authData and signature:
            rsp = {
                    0x01: credential,
                    0x02: authData,
                    0x03: signature
            }
            if userHandle:
                rsp[0x04] = {'id': userHandle}
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes(1, 'big') + cbor.dumps(rsp)
        return self._set_rsp_fields(list(result))


# Send this for every new message frame. Response contains
class CTAPHIDInitPkt(BaseStructure):
    _fields_ = [
        ('cid', 'I'),
        ('cmd', 'B'),
        ('bcnt', 'H', 0)
    ]

    def __init__(self, **kwargs):
        self.base_pack_format = '>'
        if 'data' in kwargs:
            index = None
            for i, field in enumerate(self._fields_):
                if field[0] == 'data':
                    index = i
                    break
            if index == None:
                colour_print(colour=bcolors.OKGREEN, component='CTAPHIDInitPkt.__init__', msg='setting data field')
                self._fields_ += [('data', '%ds' % len(kwargs['data']))]
            else:
                colour_print(colour=bcolors.OKGREEN, component='CTAPHIDInitPkt.__init__', 
                             msg='data already exists as a field, updating it')
                self._fields_[index] = ('data', '%ds' % len(kwargs['data']))
            logging.debug(f"{self._fields_}")
        super().__init__(**kwargs)

class CTAPHIDSeqPkt(BaseStructure):

    _fields_ = [
        ('cid', 'I'),
        ('seq', 'B'),
    ]

    def __init__(self, **kwargs):
        self.base_pack_format = '>'
        #logging.debug(kwargs)
        if 'data' in kwargs:
            index = None
            for i, field in enumerate(self._fields_):
                if field[0] == 'data':
                    index = i
                    break
            if index == None:
                logging.debug("setting data field")
                colour_print(colour=bcolors.OKPINK, component='CTAPHIDSeqPkt.__init__', msg='setting data field')
                self._fields_ += [('data', '%ds' % len(kwargs['data']))]
            else:
                colour_print(colour=bcolors.OKPINK, component='CTAPHIDSeqPkt.__init__', 
                             msg='data already exists as a field, updating it')
                self._fields_[index] = ('data', '%ds' % len(kwargs['data']))
            logging.debug(f"{self._fields_}")
        super(CTAPHIDSeqPkt, self).__init__(**kwargs)


class KeepAliveWorker(threading.Thread):

    cid = b'0xFFFFFFFF'
    not_alive = False
    uhid = None

    def __init__(self, pending, cid):
        super().__init__()
        self.pending = pending
        self.cid = cid

    def run(self):
        while self.not_alive == False:
            time.sleep(0.1)
            colour_print(colour=bcolors.FAIL, component='KeepAliveWorker.reply_with_keepalive',
                            msg='Thread reached timeout of 100ms before response buffer was recieved . . . sending heartbeat response')
            #We need to reply with keep alive after 50ms this gives us some tolerance
            #Our status is always still processing
            rsp = CTAPHIDInitPkt(cid=int.from_bytes(self.cid, 'big'),
                                  cmd=0xBB,
                                  bcnt=0x01,
                                  data=b'\x02').pack()
            self.pending.put(rsp)

    def interrupt(self):
        self.not_alive = True


class CTAP2HIDevice(UserDevice):
    #This will contain the current set of channel id's and associated state
    cids = {}

    def __init__(self, devpath):
        super().__init__(devPath=devpath)
        self.start_time = datetime.datetime.now()
        CBORCommand.set_pending(self.pending)

    def _bytes_to_str(self, b):
        return ''.join("%02X" % x for x in b)


    def send_response_segment(self, cid, cbor_cmd):
        rsp_data = None
        if cbor_cmd.response_segment == 0: #We send the init pkt
            logging.debug(f"bcnt from init pkt: {cbor_cmd.bcnt}")
            data, _ = cbor_cmd.get_rsp_seg(57)
            logging.debug(f"bcnt: {cbor_cmd.bcnt}")
            rsp_data = CTAPHIDInitPkt(cid=int.from_bytes(cid, 'big'), 
                                    cmd=cbor_cmd.ctaphid_cmd,
                                    bcnt=cbor_cmd.bcnt,
                                    data=bytes(data)).pack()
        else: #We send the continue sequence pkt
            data, seq_num = cbor_cmd.get_rsp_seg(59)
            colour_print(colour=bcolors.WARNING, component='send_response_segment', 
                    msg='Sequence number {}'.format(seq_num))
            rsp_data = CTAPHIDSeqPkt(cid=int.from_bytes(cid, 'big'),
                                     seq=seq_num,
                                     data=bytes(data)).pack()
        colour_print(colour=bcolors.WARNING, component='send_response_segment', 
                msg='pad with {} 0 bytes'.format(MAX_DATA_FRAME - len(rsp_data)))
        rsp_data += b'\00' * (MAX_DATA_FRAME - len(rsp_data)) # pad the 64 byte frame with 0x00 if required
        dump_bytes(rsp_data, colour=bcolors.OKGREEN, component='CTAP2HIDevice.send_response_segment', 
                   msg='Packed response: ')
        self.pending.put(rsp_data)


    def send_response_segments(self, cid, cbor_cmd):
        while len(cbor_cmd.response) > 0:
            self.send_response_segment(cid, cbor_cmd)
            logging.debug(f"response left: {cbor_cmd}")


    def ctaphid_ping(self, usb_req):
        cid = usb_req.data[0:4]
        cborCmd = CBORCommand(cid, None, skip_init=True)
        cborCmd.ctaphid_cmd = 0x01
        cborCmd.response = list(b'U2F_V2')
        self.send_response_segment(cid, cborCmd)

    def ctaphid_msg(self, usb_req):

        class U2FCommand(Enum):
            U2F_VERSION = 0x0
            U2F_REGISTER = 0x1
            U2F_AUTHENTICATE = 0x2
            U2F_VER = 0x03

        #Only supporting extended length encoding, section 3.1.3
        #https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.html
        cid = usb_req.data[0:4]
        if not cid in self.cids:
            colour_print(colour=bcolors.FAIL, component='CTAP2HIDevice.ctaphid_msg', 
                         msg='Unknown CID {}'.format(cid))
        cmd = usb_req.data[4:5]
        bcnt = usb_req.data[5:7]
        apdu = usb_req.data[7:]
        print("ctaphid_msg", int.from_bytes(cmd, 'big'))
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_msg', 
                    msg='cmd = {}; bcnt = {}; apdu = {}'.format(
                        self._bytes_to_str(cmd), self._bytes_to_str(bcnt), apdu))
        u2f_cla = apdu[:1]
        u2f_ins = apdu[1:2]
        u2f_p1 = apdu[2:3]
        u2f_p2 = apdu[3:4]
        u2f_lc = apdu[4:7]
        u2f_data = apdu[7:]
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_msg', 
                    msg='U2F Raw Message CLA = {}; INS = {}; P1 = {}; P2 = {}; Lc = {}'.format(
                        u2f_cla, u2f_ins, u2f_p1, u2f_p2, u2f_lc))
        if u2f_cla == b'\x00' and u2f_ins == b'\x03':
            #U2F_VERSION request, send the expected response
            cborCmd = CBORCommand(cid, None, skip_init=True)
            cborCmd.ctaphid_cmd = int.from_bytes(cmd, 'big')
            cborCmd.bcnt = 6
            cborCmd.response = list(b'U2F_V2')
            self.send_response_segment(cid, self.cids[cid]['cborCmd'])

    def ctaphid_init(self, usb_req):
        cid = usb_req.data[0:4]
        cmd = usb_req.data[4:5]
        bcnt = usb_req.data[5:7]
        nonce = usb_req.data[7:15]
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_init', 
                msg='Nonce {}'.format(self._bytes_to_str(nonce)))
        assignedCID = bytes([0, random.randint(0, 255), 0, random.randint(0, 255)])
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_init', 
                    msg='Assigning a new CID to {}'.format(self._bytes_to_str(assignedCID)))
        data = nonce + assignedCID
        # protocol == 2; major version == 5; minor version = 1; build version = 2; flags === CAPABILITY_WINK | CAPABILITY_CBOR | CAPABILITY_NMSG
        for i in [2, 5, 1, 2, 0x01 | 0x04 | 0x08]:
            data += i.to_bytes(1, 'big')
        dump_bytes(data, colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_init', msg='Response data')
        data += b'\00' * (57 - len(data)) # 64 - 4 (CID) - 1 (cmd) - 2 (bcnt) - len of response
        initCmd = CBORCommand(cid, None, skip_init=True)
        self.cids[assignedCID] = { }
        initCmd.response = data
        initCmd.ctaphid_cmd = int.from_bytes(cmd, 'big')
        initCmd.bcnt = 17
        self.send_response_segment(cid, initCmd)

    def ctaphid_cbor(self, usb_req):
        cid = usb_req.data[0:4]
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_cbor', 
                    msg='CBOR message recieved on channel {}'.format(self._bytes_to_str(cid)))
        cmd = usb_req.data[4:5]
        bcnt = usb_req.data[5:7]
        ctap_cmd = usb_req.data[7:8]
        logging.debug(f"CBOR bcnt: {int.from_bytes(bcnt, 'big') - 1}")
        cbor_data = usb_req.data[8: 7 + int.from_bytes(bcnt, 'big')]
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_cbor', 
                     msg='CBOR msg frame cmd: {}; bcnt: {}'.format(self._bytes_to_str(ctap_cmd),
                                                                   self._bytes_to_str(bcnt)))
        dump_bytes(cbor_data, colour=bcolors.OKGREEN, component='CTAP2HIDevice.ctaphid_cbor', 
                    msg='CBOR encoded bytes: ')
        cbor_cmd = CBORCommand(cid, usb_req.data[5:MAX_DATA_FRAME])
        cbor_cmd.ctaphid_cmd = int.from_bytes(cmd, 'big')
        if cbor_cmd.response_ready == True: #We can respond immediatly
            dump_bytes(cbor_cmd.response, colour=bcolors.OKGREEN, 
                       component='CTAP2HIDevice.ctaphid_cbor', msg='CBOR response: ')
            return self.send_response_segments(cid, cbor_cmd)
        else:
            self.cids[cid]['cborCmd'] = cbor_cmd
            colour_print(colour=bcolors.OKYELLOW, component='CTAP2HIDevice.ctaphid_cbor', 
                         msg="Waiting for rest of command to arrive . . .")
            return

    def _ctap_ack(self, usb_req):
        cid = usb_req.data[0:4]
        rsp = CBORCommand(cid, None, skip_init=True)
        rsp.response = []
        #if cid in self.cids:
        #    self.cids[cid]['cborCmd'] = rsp
        self.send_response_segment(cid, rsp)

    def ctaphid_cancel(self, usb_req):
        return self._ctap_ack(usb_req)

    def ctaphid_keepalive(self, usb_req):
        return

    def ctaphid_wink(self, usb_req):
        return self._ctap_ack(usb_req)

    def ctaphid_error(self, usb_req):
        return self._ctap_ack(usb_req)

    def ctaphid_unknown(self, usb_req):
        colour_print(colour=bcolors.FAIL, component='CTAP2HIDevice.ctaphid_unknown', msg='Unkown request recieved')
        self._ctap_ack(usb_req)

    def _handle_incoming_cmd(self, cmd, usb_req):
        ctapCmd = int.from_bytes(cmd, 'big') & 0x7F
        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice._handle_incoming_cmd', 
                    msg='recieved command {}'.format(ctapCmd))
        return {
            1: self.ctaphid_ping,
            3: self.ctaphid_msg,
            6: self.ctaphid_init,
            8: self.ctaphid_wink,
            16: self.ctaphid_cbor,
            17: self.ctaphid_cancel,
            59: self.ctaphid_keepalive,
            63: self.ctaphid_error,
        }.get(ctapCmd, self.ctaphid_unknown)(usb_req)

    def _handle_incoming_sequence(self, cid, usb_req):
        seqNum = int.from_bytes(usb_req.data[4:5], 'big')
        context = self.cids.get(cid)
        if context != None:
            transaction = context.get("cborCmd")
            if transaction != None and seqNum == transaction.request_segment:
                transaction.append_segment(usb_req.data[5:MAX_DATA_FRAME])
                if transaction.response_ready == True:
                    del self.cids[cid]['cborCmd']
                    self.send_response_segments(cid, transaction)
                else:
                    colour_print(colour=bcolors.OKPURPLE, component='CTAP2HIDevice._handle_incoming_sequence', 
                                 msg='Sequence number [{}] not the last expected sequence'.format(seqNum))
            else:
                colour_print(colour=bcolors.FAIL, component='CTAP2HIDevice._handle_incoming_sequence', 
                             msg='Sequence number [{}] not the next expected sequence [{}]'.format(
                                 seqNum, transaction.request_segment))
        else:
            colour_print(colour=bcolors.FAIL, component='CTAP2HIDevice._handle_incoming_sequence', 
                         msg='CID not found in device context, don\'t know what to do')

    def process_output(self, event):
        ep = event.data[0]
        event.data = event.data[1:]
        cid = event.data[0:4]
        cmd = event.data[4:5]
        dump_bytes(event.data[:event.ev_len])

        colour_print(colour=bcolors.OKGREEN, component='CTAP2HIDevice._handle_incoming', 
                    msg='EP: {} CID: {}; CMD/SEQ {}; DATA: {}'.format(
                        ep, cid, cmd, self._bytes_to_str(event.data[:event.ev_len])))

        if(int.from_bytes(cmd, 'big') & 0x80) > 0:
            colour_print(colour=bcolors.FAIL, component='CTAP2HIDevice._handle_incoming', 
                        msg='bit 8 set we got a command msg')
            return self._handle_incoming_cmd(cmd, event)
        else:
            colour_print(colour=bcolors.OKPURPLE, component='CTAP2HIDevice._handle_incoming',
                         msg='Recieved a sequence segment, appending it to the current msg context')
            return self._handle_incoming_sequence(cid, event)

    def join(self, timeout=None):
        super().join()
        AuthenticatorAPI.quit()



'''
Copyright (c) 2014 Yaron Shani <yaron.shani@gmail.com>.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

This software is provided ``as is'' and any express or implied
warranties, including, but not limited to, the implied warranties of
merchantability and fitness for a particular purpose are
disclaimed. In no event shall author or contributors be liable for any
direct, indirect, incidental, special, exemplary, or consequential
damages (including, but not limited to, procurement of substitute
goods or services; loss of use, data, or profits; or business
interruption) however caused and on any theory of liability, whether
in contract, strict liability, or tort (including negligence or
otherwise) arising in any way out of the use of this software, even if
advised of the possibility of such damage.

Update 2022 by Lachlan Gleeson for python 3
'''

import socketserver, datetime, struct, traceback, re, signal


# Hey StackOverflow !
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    OKPINK = '\033[95m'
    OKYELLOW = '\033[93m'
    OKPURPLE = '\033[35m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def colour_print(colour=bcolors.OKBLUE, component='USB/IP', msg=''):
    print('[' + colour + component + bcolors.ENDC + '] ' + msg)


def print_bytes(*args):
    result = ""
    count = 0
    for ba in args:
        for x in ba:
            result += "%02X " % x
            count += 1
            if count == 8 :
                result += " "
            elif count == 16:
                print("\t" + result)
                result = ""
                count = 0
    print('\t' + result + '\n')

def dump_bytes(*args, colour=bcolors.OKPURPLE, component='USB/IP CONTROLLER', msg=''):
    #Print bytes in nice format
    c = colour if colour != None else bcolors.OKPURPLE
    colour_print(colour=colour, component=component, msg=msg)
    print_bytes(*args)


class BaseStructure(object):
    _fields_ = []

    def __init__(self, **kwargs):
        self.init_from_dict(**kwargs)
        for field in self._fields_:
            if len(field) > 2:
                if not hasattr(self, field[0]):
                    setattr(self, field[0], field[2])

    def init_from_dict(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def size(self):
        return struct.calcsize(self.format())

    def format(self):
        pack_format = '>'
        for field in self._fields_:
            if isinstance(field[1], BaseStructure):
                pack_format += str(field[1].size()) + 's'
            elif 'si' == field[1]:
                pack_format += 'c'
            elif '<' in field[1] or '>' in field[1]:
                pack_format += field[1][1:]
            else:
                pack_format += field[1]
        print(pack_format)
        return pack_format.encode('utf-8')

    def formatDevicesList(self, devicesCount):

        pack_format = '>'
        i = 0
        for field in self._fields_:
            if (i == devicesCount + 2):
                break
            if isinstance(field[1], BaseStructure):
                pack_format += str(field[1].size()) + 's'
            elif 'si' == field[1]:
                pack_format += 'c'
            elif '<' in field[1]:
                pack_format += field[1][1:]
            else:
                pack_format += field[1]
            i += 1
        #print(pack_format)
        return pack_format.encode('utf-8')

    def pack(self):
        values = []
        for field in self._fields_:
            #print("Field: {}".format(field))
            if isinstance(field[1], BaseStructure):
                values.append(getattr(self, field[0], 0).pack())
            elif re.match(r'\d*x', field[1]):
                #Skipp padding
                continue
            else:
                if 'si' == field[1]:
                    values.append(chr(getattr(self, field[0], 0)))
                else:
                    values.append(getattr(self, field[0], 0))
        #Python 2 -> 3, str != bytestring so conditionally remap any strings we find.
        values = [bytes(v, 'utf-8') if isinstance(v, str) else v for v in values]
        #print(values)
        packed = struct.pack(self.format(), *values)
        #print("packed [{}]".format(packed))
        return packed 

    def packDevicesList(self, devicesCount):
        values = []
        i = 0
        for field in self._fields_:
            if (i == devicesCount + 2):
                break
            if isinstance(field[1], BaseStructure):
                values.append(getattr(self, field[0], 0).pack())
            else:
                if 'si' == field[1]:
                    values.append(chr(getattr(self, field[0], 0)))
                else:
                    values.append(getattr(self, field[0], 0))
            i += 1
        return struct.pack(self.formatDevicesList(devicesCount), *values)

    def unpack(self, buf):
        values = struct.unpack(self.format(), buf)
        i=0
        keys_vals = {}
        for val in values:
            if '<' in self._fields_[i][1][0]:
                val = struct.unpack('<' +self._fields_[i][1][1], struct.pack('>' + self._fields_[i][1][1], val))[0]
            keys_vals[self._fields_[i][0]]=val
            i+=1
        #print(keys_vals)
        self.init_from_dict(**keys_vals)


class USBIPHeader(BaseStructure):
    _fields_ = [
        ('version', 'H', 273),
        ('command', 'H'),
        ('status', 'I')
    ]


class USBInterface(BaseStructure):
    _fields_ = [
        ('bInterfaceClass', 'B'),
        ('bInterfaceSubClass', 'B'),
        ('bInterfaceProtocol', 'B'),
        ('align', 'B', 0)
    ]

class USBIPDevice(BaseStructure):
    _fields_ = [
        ('usbPath', '256s'),
        ('busID', '32s'),
        ('busnum', 'I'),
        ('devnum', 'I'),
        ('speed', 'I'),
        ('idVendor', 'H'),
        ('idProduct', 'H'),
        ('bcdDevice', 'H'),
        ('bDeviceClass', 'B'),
        ('bDeviceSubClass', 'B'),
        ('bDeviceProtocol', 'B'),
        ('bConfigurationValue', 'B'),
        ('bNumConfigurations', 'B'),
        ('bNumInterfaces', 'B'),
        ('interfaces', USBInterface())
    ]

class OPREPDevList(BaseStructure):

    def __init__(self, dictArg, count):
        self._fields_ = [
            ('base', USBIPHeader(), USBIPHeader(command=5,status=0)), # Declare this here to make sure it's in the right order
            ('nExportedDevice', 'I', count) # Same for this guy
        ]

        for key, value in dictArg.items():
            field = (str(key), value[0], value[1])
            self._fields_.append(field)

        for field in self._fields_:
            if len(field) > 2:
                if not hasattr(self, field[0]):
                    setattr(self, field[0], field[2])

class OPREPImport(BaseStructure):
    _fields_ = [
        ('base', USBIPHeader()),
        ('usbPath', '256s'),
        ('busID', '32s'),
        ('busnum', 'I'),
        ('devnum', 'I'),
        ('speed', 'I'),
        ('idVendor', 'H'),
        ('idProduct', 'H'),
        ('bcdDevice', 'H'),
        ('bDeviceClass', 'B'),
        ('bDeviceSubClass', 'B'),
        ('bDeviceProtocol', 'B'),
        ('bConfigurationValue', 'B'),
        ('bNumConfigurations', 'B'),
        ('bNumInterfaces', 'B')
    ]

# https://www.kernel.org/doc/html/v5.14/usb/usbip_protocol.html

class USBIPRETSubmit(BaseStructure):

    '''
    def __init__(self, **kwargs):
        if 'data_frame' in kwargs:
            self._fields_ += [('data_frame', "%ds" % len(kwargs['data_frame']))]
            print(self._fields_)
        super(USBIPRETSubmit, self).__init__(**kwargs)
    '''

    _fields_ = [
        ('command', 'I'),
        ('seqnum', 'I'),
        ('devid', 'I'),
        ('direction', 'I'),
        ('ep', 'I'),
        ('status', 'I'),
        ('actual_length', 'I'),
        ('start_frame', 'I'),
        ('number_of_packets', 'I'),
        ('error_count', 'I'),
        ('padding', 'Q')
    ]

    def pack(self):
        packed_data = BaseStructure.pack(self)
        #print("packed_data: [{}]".format(packed_data))
        #print("self.data: [{}]".format(self.data))
        if isinstance(self.data_frame, str):
            self.data_frame = self.data_frame.encode()
        packed_data += self.data_frame
        return packed_data

class USBIPCMDUnlink(BaseStructure):
    _fields_ = [
        ('seqnum', 'I'),
        ('devid', 'I'),
        ('direction', 'I'),
        ('ep', 'I'),
        ('seqnum2', 'I'),
    ]

'''
class USBIBCMDBasic(BaseStructure):
    _fields_ = [
        ('command', 'I'), #0x1
        ('seqnum', 'I'),
        ('devid', 'I'),
        ('direction', 'I'),
        ('ep', 'I')
    ]
'''

class USBIPCMDSubmit(BaseStructure):
    _fields_ = [
        ('command', 'I'),
        ('seqnum', 'I'),
        ('devid', 'I'),
        ('direction', 'I'),
        ('ep', 'I'),
        ('transfer_flags', 'I'),
        ('transfer_buffer_length', 'I'),
        ('start_frame', 'I'),
        ('number_of_packets', 'I'),
        ('interval', 'I'),
        ('setup', 'Q')
    ]

class USBIPUnlinkReq(BaseStructure):
    _fields_ = [
        ('seqnum', 'I'),
        ('devid', 'I'),
        ('direction', 'I'),
        ('ep', 'I'),
        ('unlink_seqnum', 'I'),
        ('padding', '24x')
    ]

class USBIPUnlinkRet(BaseStructure):
    _fields_ = [
        ('command', 'I', 0x4),
        ('seqnum', 'I'),
        ('devid', 'I', 0x2),
        ('direction', 'I'),
        ('ep', 'I'),
        ('status', 'I'),
        ('padding', '24x')
    ]


class StandardDeviceRequest(BaseStructure):
    _fields_ = [
        ('bmRequestType', 'B'),
        ('bRequest', 'B'),
        ('wValue', 'H'),
        ('wIndex', 'H'),
        ('wLength', '<H')
    ]

class DeviceDescriptor(BaseStructure):
    _fields_ = [
        ('bLength', 'B', 18),
        ('bDescriptorType', 'B', 1),
        ('bcdUSB', 'H', 0x1001),
        ('bDeviceClass', 'B'),
        ('bDeviceSubClass', 'B'),
        ('bDeviceProtocol', 'B'),
        ('bMaxPacketSize0', 'B'),
        ('idVendor', '>H'),
        ('idProduct', '>H'),
        ('bcdDevice', 'H'),
        ('iManufacturer', 'B'),
        ('iProduct', 'B'),
        ('iSerialNumber', 'B'),
        ('bNumConfigurations', 'B')
    ]

class DeviceConfigurations(BaseStructure):
    _fields_ = [
        ('bLength', 'B', 9),
        ('bDescriptorType', 'B', 2),
        ('wTotalLength', 'H', 0x2900),
        ('bNumInterfaces', 'B', 1),
        ('bConfigurationValue', 'B', 1),
        ('iConfiguration', 'B', 0),
        ('bmAttributes', 'B', 0x80),
        ('bMaxPower', 'B', 0x32)
    ]


class InterfaceDescriptor(BaseStructure):
    _fields_ = [
        ('bLength', 'B', 9),
        ('bDescriptorType', 'B', 4),
        ('bInterfaceNumber', 'B', 0),
        ('bAlternateSetting', 'B', 0),
        ('bNumEndpoints', 'B', 1),
        ('bInterfaceClass', 'B', 3),
        ('bInterfaceSubClass', 'B', 1),
        ('bInterfaceProtocol', 'B', 2),
        ('iInterface', 'B', 0)
    ]


class EndPoint(BaseStructure):
    _fields_ = [
        ('bLength', 'B', 7),
        ('bDescriptorType', 'B', 0x5),
        ('bEndpointAddress', 'B', 0x81),
        ('bmAttributes', 'B', 0x3),
        ('wMaxPacketSize', 'H', 0x8000),
        ('bInterval', 'B', 0x0A)
    ]



class USBRequest():
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class USBDevice():
    '''interfaces = [USBInterface(bInterfaceClass=0x3, bInterfaceSubClass=0x0, bInterfaceProtocol=0x0)]
    speed=2
    speed = 2
    vendorID = 0xc410
    productID = 0x0
    bcdDevice = 0x0
    bDeviceClass = 0x0
    bDeviceSubClass = 0x0
    bDeviceProtocol = 0x0
    bNumConfigurations = 1
    bConfigurationValue = 1
    bNumInterfaces = 1'''

    def __init__(self):
        self.generate_raw_configuration()
        self.start_time = datetime.datetime.now()

    def generate_raw_configuration(self):
        _str = self.configurations[0].pack()
        _str += self.configurations[0].interfaces[0].pack()
        _str += self.configurations[0].interfaces[0].descriptions[0].pack()
        #_str += self.configurations[0].interfaces[0].endpoints[0].pack()
        for e in self.configurations[0].interfaces[0].endpoints:
            _str += e.pack()
        self.all_configurations = _str


    def send_usb_req(self, usb_res, usb_len, status=0, ep=0, start_frame=0, packets=0, seqnum=None, 
                     direction=0):
        rsp = USBIPRETSubmit(command=0x3,
                             seqnum=seqnum,
                             devid=0,
                             direction=direction,
                             ep=ep,
                             status=status,
                             actual_length=usb_len,
                             start_frame=start_frame,
                             number_of_packets=packets,
                             error_count=0,
                             interval=0x0,
                             padding=0,
                             data_frame=usb_res).pack()
        dump_bytes(list(rsp), colour=bcolors.FAIL, component='USBDevice(response)', msg='response bytes:')
        self.connection.sendall(rsp)

    def handle_get_descriptor(self, control_req, usb_req):
        handled = False
        #print("handle_get_descriptor {}".format(control_req.wValue,'n'))
        if control_req.wValue == 0x1: # Device
            handled = True
            ret=DeviceDescriptor(bDeviceClass=self.bDeviceClass,
                                 bDeviceSubClass=self.bDeviceSubClass,
                                 bDeviceProtocol=self.bDeviceProtocol,
                                 bMaxPacketSize0=8,
                                 idVendor=self.vendorID,
                                 idProduct=self.productID,
                                 bcdDevice=self.bcdDevice,
                                 iManufacturer=0,
                                 iProduct=0,
                                 iSerialNumber=0,
                                 bNumConfigurations=1).pack()
            self.send_usb_req(ret, len(ret), seqnum=usb_req.seqnum)
        elif control_req.wValue == 0x2: # configuration descriptor
            handled = True
            ret= self.all_configurations[:control_req.wLength]
            #print(ret)
            self.send_usb_req(ret, len(ret), seqnum=usb_req.seqnum)
        return handled


    def handle_set_configuration(self, control_req, usb_req):
        handled = True
        self.send_usb_req(b'', 0, seqnum=usb_req.seqnum)
        return handled

    def handle_usb_control(self, usb_req):
        control_req = StandardDeviceRequest()
        control_req.unpack(usb_req.setup.to_bytes(8, 'big'))
        handled = False
        print('[' + bcolors.OKBLUE + 'USBDevice(handle_usb_control)' + bcolors.ENDC + "] UC Request Type" + \
                " {}; UC Request {}; UC Value  {}; UCIndex  {}; UC Length {}".format(
                control_req.bmRequestType, control_req.bRequest, control_req.wValue, control_req.wIndex,
                control_req.wLength))
        if control_req.bmRequestType == 0x80: # Host Request
            if control_req.bRequest == 0x06: # Get Descriptor
                handled = self.handle_get_descriptor(control_req, usb_req)
            if control_req.bRequest == 0x00: # Get STATUS
                self.send_usb_req(b"\x01\x00", 2, seqnum=usb_req.seqnum);
                handled = True

        if control_req.bmRequestType == 0x00: # Host Request
            if control_req.bRequest == 0x09: # Set Configuration
                handled = self.handle_set_configuration(control_req, usb_req)
        if not handled:
            self.handle_unknown_control(control_req, usb_req)

    def handle_usb_request(self, usb_req):
        try:
            if usb_req.ep == 0:
                print('[' + bcolors.OKBLUE + 'USBDevice(handle_usb_request)' + bcolors.ENDC + '] Control request')
                self.handle_usb_control(usb_req)
            else:
                print('[' + bcolors.OKBLUE + 'USBDevice(handle_usb_request)' + bcolors.ENDC + \
                        '] Data request for ep {}'.format(usb_req.ep))
                self.handle_data(usb_req)
        except Exception as e:
            print(e)
            traceback.print_exc()
            raise e

class USBContainer:
    usb_devices = {}
    attached_devices = {}
    devices_count = 0

    def add_usb_device(self, usb_device):
        self.devices_count += 1
        busID = '1-1.' + str(self.devices_count)
        self.usb_devices[busID] = usb_device
        self.attached_devices[busID] = False

    def remove_usb_device(self, usb_device):
        for busid, dev in self.usb_devices.items():
            if dev == usb_device:
                del self.attached_devices[busid]
                del self.usb_devices[busid]
                break
        self.devices_count -= 1

    def detach_all(self):
        self.attached_devices = {}
        self.usb_devices = {}
        self.devices_count = 0

    def handle_attach(self, busid):
        if (self.usb_devices[busid] != None):
            busnum = int(busid[4:])
            return OPREPImport(base=USBIPHeader(command=3, status=0),
                               usbPath='/sys/devices/pci0000:00/0000:00:01.2/usb1/' + busid,
                               busID=busid,
                               busnum=busnum,
                               devnum=2,
                               speed=2,
                               idVendor=self.usb_devices[busid].vendorID,
                               idProduct=self.usb_devices[busid].productID,
                               bcdDevice=self.usb_devices[busid].bcdDevice,
                               bDeviceClass=self.usb_devices[busid].bDeviceClass,
                               bDeviceSubClass=self.usb_devices[busid].bDeviceSubClass,
                               bDeviceProtocol=self.usb_devices[busid].bDeviceProtocol,
                               bNumConfigurations=self.usb_devices[busid].bNumConfigurations,
                               bConfigurationValue=self.usb_devices[busid].bConfigurationValue,
                               bNumInterfaces=self.usb_devices[busid].bNumInterfaces)

    def handle_device_list(self):
        devices = {}

        i = 0
        for busid, usb_dev in self.usb_devices.items():
            i += 1
            devices['device' + str(i)] = [USBIPDevice(), USBIPDevice(
                usbPath='/sys/devices/pci0000:00/0000:00:01.2/usb1/' + busid,
                busID=busid,
                busnum=i,
                devnum=2,
                speed=2,
                idVendor=self.usb_devices[busid].vendorID,
                idProduct=self.usb_devices[busid].productID,
                bcdDevice=self.usb_devices[busid].bcdDevice,
                bDeviceClass=self.usb_devices[busid].bDeviceClass,
                bDeviceSubClass=self.usb_devices[busid].bDeviceSubClass,
                bDeviceProtocol=self.usb_devices[busid].bDeviceProtocol,
                bNumConfigurations=self.usb_devices[busid].bNumConfigurations,
                bConfigurationValue=self.usb_devices[busid].bConfigurationValue,
                bNumInterfaces=self.usb_devices[busid].bNumInterfaces,
                interfaces=USBInterface(bInterfaceClass=self.usb_devices[busid].configurations[0].interfaces[0].bInterfaceClass,
                                        bInterfaceSubClass=self.usb_devices[busid].configurations[0].interfaces[0].bInterfaceSubClass,
                                        bInterfaceProtocol=self.usb_devices[busid].configurations[0].interfaces[0].bInterfaceProtocol)
            )]

        return OPREPDevList(devices, len(self.usb_devices))


    def run(self, ip='0.0.0.0', port=3240):
        colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='Starting server')
        socketserver.TCPServer.allow_reuse_address = True
        self.server = socketserver.ThreadingTCPServer((ip, port), USBIPConnection)
        self.server.usbcontainer = self
        self.server.serve_forever()


class USBIPConnection(socketserver.BaseRequestHandler):
    attached = False
    attachedBusID = ''

    def __init__(self, request=None, client_address=None, server=None):
        super().__init__(request=request, client_address=client_address, server=server)
        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.interrupt)

    def interrupt():
        if self.request:
            self.request.close()
        if self.server:
            self.server.server_close()

    def handle(self):
        endpoint_requests = {}
        colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='New connection from {}'.format(self.client_address))
        req = USBIPHeader()
        while 1:
            if not self.attached:
                data = self.request.recv(8)
                if not data:
                    break
                req.unpack(data)
                colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='Header packet is valid')
                colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='Command is {}'.format(hex(req.command)))
                if req.command == 0x8005:
                    colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='Querying device list')
                    self.request.sendall(self.server.usbcontainer.handle_device_list().pack())
                elif req.command == 0x8003:
                    busid = self.request.recv(5).strip()  # receive bus id
                    colour_print(colour=bcolors.OKBLUE, component='USBIP', 
                                 msg='Attaching to device with busid [{}]'.format(busid.decode()))
                    self.request.recv(27)
                    self.request.sendall(self.server.usbcontainer.handle_attach(busid.decode()).pack())
                    self.attached = True
                    self.attachedBusID = busid.decode()
                    colour_print(colour=bcolors.OKBLUE, component='USBIP', msg='attached')

            else:
                #print(self.server.usbcontainer.usb_devices)
                if (not self.attachedBusID in self.server.usbcontainer.usb_devices):
                    colour_print(colour=bcolors.WARNING, component='USBIP', msg='closing')
                    self.request.close()
                    break
                else:
                    colour_print(component='USB/IP', msg='waiting for command')
                    command = self.request.recv(4)
                    dump_bytes(command, msg='USB/IP command bytes recieved:')
                    cmdVal = struct.unpack('>I', command)[0]
                    '''
                    if (cmdVal == 0x00000003):
                        cmd = USBIPCMDUnlink()
                        data = self.request.recv(cmd.size())
                        cmd.unpack(data)
                        colour_print(component='USBIP', msg='Detaching device with seqnum {}'.format(cmd.seqnum))
                        # We probably don't even need to handle that, the windows client doesn't even send this packet
                    '''
                    if (cmdVal == 0x00000001):
                        cmd = USBIPCMDSubmit()
                        data = self.request.recv(cmd.size() - 4)
                        cmd.unpack(command + data)
                        msg = 'USB/IP Command::\n\tseqnum: {}; devid: {};\n\tdirection: {}; ep: {};\n\tflags: {};'\
                                'transfer buffer: {};\n\tstart_frame: {}; no. of pkts: {}; '\
                                '\n\tinterval: {}; setup: {}'.format(
                            cmd.seqnum,cmd.devid,cmd.direction,cmd.ep,cmd.transfer_flags,cmd.transfer_buffer_length,
                            cmd.start_frame,cmd.number_of_packets,cmd.interval,list(cmd.setup.to_bytes(8, 'big')))
                        colour_print(colour=bcolors.OKBLUE, component='USBIPConnection.handle', msg=msg)
                        if endpoint_requests.get(cmd.ep) == None:
                            endpoint_requests[cmd.ep] = 1
                        else:
                            endpoint_requests[cmd.ep] = endpoint_requests.get(cmd.ep) + 1
                        msg = "Endpoint requests: {}".format( endpoint_requests)
                        colour_print(component='USBIPConnection.handle', msg=msg)
                        data_frame = b''
                        if cmd.start_frame == 0xFFFFFFFF and cmd.transfer_flags == 0x0:
                            colour_print(colour=bcolors.OKYELLOW, component='USBIPConnection.handle', msg='CTAPHID:: '\
                                    'FIDO2 Authenticator recieved start_frame, reading rest of data maybe . . .')
                            data_frame = self.request.recv(cmd.transfer_buffer_length)
                            dump_bytes(data_frame, component='USBIPConnection.handle', msg='data bytes recieved:')
                        usb_req = USBRequest(seqnum=cmd.seqnum,
                                             devid=cmd.devid,
                                             direction=cmd.direction,
                                             ep=cmd.ep,
                                             flags=cmd.transfer_flags,
                                             number_of_packets=cmd.number_of_packets,
                                             interval=cmd.interval,
                                             setup=cmd.setup,
                                             cmd_frame=cmd.pack(),
                                             data_frame=data_frame)
                        dump_bytes(list(usb_req.setup.to_bytes(8, 'big')), colour=bcolors.FAIL, 
                                   component='USBDevice(send_usb_req)', msg='setup bytes:')
                        dump_bytes(list(usb_req.cmd_frame), list(usb_req.data_frame), colour=bcolors.FAIL, 
                                    component='USBDevice(request)', msg='whole recieved message:')
                        self.server.usbcontainer.usb_devices[self.attachedBusID].connection = self.request
                        try:
                            self.server.usbcontainer.usb_devices[self.attachedBusID].handle_usb_request(usb_req)
                        except:
                            colour_print(colour=bcolors.FAIL, component='USBIP', 
                                         msg='Connection with client ' + str(self.client_address) + ' ended')
                            break
                    elif(cmdVal == 0x00000002):
                        cmd = USBIPUnlinkReq()
                        data = self.request.recv(cmd.size())
                        cmd.unpack(data)
                        dump_bytes(command + cmd.pack(), colour=bcolors.WARNING, component='USBIP', msg='Unlink request')
                        #TODO have we actually sent a USBIP_RET_SUBMIT or not?
                        success = self.server.usbcontainer.usb_devices[self.attachedBusID].unlink(cmd)
                        status = 0x0;
                        if success == True:
                            status = 0xF
                        ret = USBIPUnlinkRet(command=0x04, seqnum=cmd.seqnum, devid=cmd.devid, direction=0, ep=cmd.ep,
                                             status=status, padding=b'\0' * 24)
                        dump_bytes(ret.pack(), colour=bcolors.WARNING, component='USBIP', msg='Unlink return')
                        self.request.sendall(ret.pack())

                    else:
                        raise Exception("Unknown USB/IP command recieved")
        self.request.close()
        self.server.server_close()



# Copyrite IBM 2022, 2025
# IBM Confidential

import logging
import struct
import os
import cbor2 as cbor
import secrets
import base64

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, utils
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from soft_fido2.cert_utils import CertUtils



class KeyUtils(object):

    @classmethod
    def get_passkey_seed(cls, entropy, key):
        """
        Generate a 32 byte seed from a domain and a private key.

        Entropy is typically the bytes of the rp.id. Key is an Elliptic Curve key.

        Entropy is hashed using SHA256 before signing.

        Returned bytestring is b64_url encoded.
        """
        if not isinstance(entropy, bytes):
            raise ValueError(f"Entropy must be bytes: {entropy}")
        if not isinstance(key, ec.EllipticCurvePrivateKey):
            raise ValueError(f"Key must be an EllipticCurvePrivateKey: {key}")
        theHash = hashes.SHA256()
        digester = hashes.Hash(theHash)
        digester.update(entropy)
        sig = key.sign(digester.finalize(),
                        ec.ECDSA(utils.Prehashed(theHash), deterministic_signing=True))
        digester = hashes.Hash(theHash)
        digester.update(sig)
        result = base64.urlsafe_b64encode(digester.finalize()[:32])
        #logging.debug(f"start: {entropy}; sig: {sig}; return {result}")
        return result


    @classmethod
    def _long_to_bytes(cls, l):
        limit = 256**4 - 1  #max value we can fit into a struct.pack
        parts = []
        while l:
            parts.append(l & limit)
            l >>= 32
        parts = parts[::-1]
        return struct.pack(">" + 'L' * len(parts), *parts)


    @classmethod
    def _bytes_to_long(cls, b):
        l = int(len(b) / 4)
        parts = struct.unpack(">" + 'L' * l, b)[::-1]
        result = 0
        for i in range(len(parts)):
            temp = parts[i] << (32 * i)
            result += temp

        return result

    @classmethod
    def load_der_key(cls, key, secret=None):
        '''
        load DER encoded key, returns KeyPair
        '''
        pk = serialization.load_der_private_key(key, 
                password=None,
                backend=default_backend())
        return KeyPair(pk, pk.public_key())

    @classmethod
    def load_mldsa_key(cls, alg, seed):
        from oqs.oqs import Signature
        ml_key: Signature = Signature(alg, seed)
        pubkey: bytes = ml_key.generate_keypair_seed()
        return KeyPair(ml_key, pubkey)

    @classmethod
    def der_enc_key(cls, pk):
        return pk.private_bytes(encoding=serialization.Encoding.DER,
                                format=serialization.PrivateFormat.PKCS8,
                                encryption_algorithm=serialization.NoEncryption())

    @classmethod
    def get_alg_id_from_pubkey_and_hash(cls, publicKey, alg, eckx=False, pss=False):
        if isinstance(publicKey, rsa.RSAPublicKey):
            algName = alg.name if alg and hasattr(alg, 'name') else 'unknown'
            if pss:
                return {
                    'sha256': -37,
                    'sha384': -38,
                    'sha512': -39,
                }.get(alg, 0)
            return {
                'sha1': -65535,
                'sha256': -257,
                'sha384': -258,
                'sha512': -259,
            }.get(algName, 0)
        elif isinstance(publicKey, ec.EllipticCurvePublicKey):
            if isinstance(alg, hashes.SHA256):
                return -7 if eckx == False else -25
            if isinstance(alg, hashes.SHA384):
                return -35
            elif isinstance(alg, hashes.SHA512):
                return -36 if eckx == False else -26
        elif isinstance(publicKey, ed25519.Ed25519PublicKey):
            return -8
        elif isinstance(publicKey, bytes): #TODO guessing poorly supported PQC
            #Guess the type based on the length?
            #print(f"bytes[{len(publicKey)}]: {publicKey}")
            return {
                1312: -48, # Draft ML-DSA-44 with SHA256?
                1952: -49, # Draft ML-DSA-65 with SHA256?
                2592: -50 # Draft ML-DSA-87 with SHA256?
            }.get(len(publicKey), 9001)
        return 0

    @classmethod
    def get_cose_key(cls, publicKey, alg, eckx=False):
        '''
        COSE key representation of the public key
        :param publicKey: public key interface
        :param alg: hashing algorithm used
        :param eckx: True if key is used for Elliptic Curve Key Exchange (modifies key type|kty)
        :return:
        '''
        if isinstance(publicKey, rsa.RSAPublicKey):
            return {1: 3,
                    3: cls.get_alg_id_from_pubkey_and_hash(publicKey, alg),
                   -1: cls._long_to_bytes(publicKey.public_numbers().n),
                   -2: cls._long_to_bytes(publicKey.public_numbers().e)
                 }
        elif isinstance(publicKey, ec.EllipticCurvePublicKey):
            return {1: 2,
                    3: cls.get_alg_id_from_pubkey_and_hash(publicKey, alg, eckx),
                    -1: 1,
                    -2: cls._long_to_bytes(publicKey.public_numbers().x),
                    -3: cls._long_to_bytes(publicKey.public_numbers().y)
                }
        elif isinstance(publicKey, ed25519.Ed25519PublicKey):
            return {1: 6,
                    3: cls.get_alg_id_from_pubkey_and_hash(publicKey, alg),
                   -1: 6,
                   -2: publicKey.public_bytes(encoding=serialization.Encoding.Raw,
                                                  format=serialization.PublicFormat.Raw)
                 }
        elif isinstance(publicKey, bytes): #guess poorly supported PQC pubkey
            return {1: 7,
                    3: cls.get_alg_id_from_pubkey_and_hash(publicKey, alg),
                   -1: publicKey,
            }
        else:
            raise Exception("Unsupported public key algorithm")

    @classmethod
    def generate_passkey(cls):
        '''
        Generate the data required for a passkey capable of
        packed attestation with a claimed aaguid.

        '''       
        # Generate key pair
        kp = KeyPair.generate_ecdsa()
        
        # Generate certificate
        subj = x509.Name([
            x509.NameAttribute(x509.NameOID.COMMON_NAME, u'Pirate Passkey'),
            x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'EyeBeeKey')
        ])
        pem = CertUtils.gen_ca_cert(subject=subj, lifetime=9999, keyPair=kp)
        
        # Create passkey data
        passkey_data = {
            'x5c': pem,
            'key': kp.get_private(),
            #seed === sha256 of rp.id signed by key
        }
        return passkey_data

    @classmethod
    def update_passkey(cls, resCred, pinHash, passkeyFilename):
        '''
        Add a resident cred to a .passkey file
        '''
        passkey = cls._load_passkey(pinHash, passkeyFilename)
        res_creds = [ *passkey.get('res.creds', []), resCred]
        cls._save_passkey(
            passkey['key'],
            passkey['x5c'],
            res_creds,
            passkey['pin.hash'],
            passkeyFilename
        )

    '''
    Passkey File:
        key: EC key
        x5c: X509 certificate issued to key
        res.creds: list of resident credentials (dictionary of "rp.id", "user.id", "cred.id")
        pin.hash: SHA256 of user provided secret, only the lower half (16 bytes) of this value
                  is provided during pin auth protocol 1

        File: Header | Body
        Header: enc.upper.hash: bytes(230)
        Body: pkcs12.len: bytes(4) | pkcs12.file: bytes(pcks12.len) | enc.res.creds: bytes(remaining)

    Write file process:
        upper.hash: bytes 16-32 of the full 32-byte hash of pin (pin.hash)
        enc.upper.hash: ec_encrypt upper.hash with ${FIDO_HOME}/platform.key
        header: enc.upper.hash
        pkcs12.bytes: use pin.hash as secret to generate encrypted pkcs12 bytes of key + x5c
        pkcs12.bytes.len: len(pcks12.bytes)
        enc.res.creds : encrypt cbor.encode(res_creds) with key
        body: concatenate pcks12.bytes.len | pkcs12.bytes | enc.res.creds
        file: concatenate header | body
        write file to disk

    Read file process:
        collect pin (lower pin hash for pin auth protocol, ect,)
        read file, split header from body
        upper.hash: ec_decrypt enc.upper.hash with ${FIDO_HOME}/platform.key
        pin.hash: concatenate lower.hash | upper.hash
        read pkcs12.bytes.len
        read pkcs12.bytes 
        key | x5c: use pin hash to decrypt pcks12 file
        enc.res.creds: read remaining
        cbor.res.creds: decrypt enc.res.creds with key
        res.creds: cbor.decode(cbor.res.creds)

    '''

    @classmethod
    def __read_passkey(cls, passkeyFilename):
        if not passkeyFilename.endswith('.passkey'):
            passkeyFilename += '.passkey'
        passkey_path = os.path.join(os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido')), passkeyFilename)
        with open(passkey_path, 'rb') as f:
            # Read the entire file
            file_content = f.read()
            
            # Split header from body
            # Header is the encrypted upper hash (32 bytes)
            header = file_content[:230]
            body = file_content[230:]
        return header, body

    @classmethod
    def __get_upper_hash(cls, ciphertext, secret):
        # Get platform key to decrypt the header
        platform_key_path = os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido')) + '/platform.key'
        with open(platform_key_path, 'rb') as key_file:
            platform_key_pem = key_file.read()
            platform_key = serialization.load_pem_private_key(
                platform_key_pem,
                password=secret,
                backend=default_backend()
            ) 
        # Decrypt the upper hash using the platform key
        return cls.ec_decrypt(ciphertext, platform_key)

    @classmethod
    def _load_passkey(cls, pinHash, passkeyFilename, secret=None):
        """
        Read passkey file according to the process described in the comments.
        
        Args:
            pinHash: SHA256 hash of the user's PIN (lower half for pin auth protocol)
            passkeyFilename: Path to the passkey file
            secret [optional]: Secret used to open platform key.
            
        Returns:
            Dictionary containing key, x5c, and res.creds
        """
        # We expect pinHash to be the lower half (16 bytes) used for pin auth protocol
        header, body = cls.__read_passkey(passkeyFilename)
        passkey = {}
        pin_hash = pinHash
        if len(pinHash) == 16:
            # Reconstruct the full pin hash
            pin_hash = pinHash + cls.__get_upper_hash(header, secret) 

        # Read PKCS12
        pkcs12_len = int.from_bytes(body[:4], 'little')
        pkcs12_bytes = body[4:4+pkcs12_len]
        key, certificate = cls.load_key_and_cert(pkcs12_bytes,
                                                      base64.b64encode(pin_hash))

        res_creds = []
        if len(body) > (pkcs12_len + 4):
            cbor_res_creds = cls.ec_decrypt(body[4+pkcs12_len:], key.get_private())
            res_creds = cbor.loads(cbor_res_creds)
            if not isinstance(res_creds, list):
                raise ValueError('res_creds is not a list')
            elif len(res_creds) > 0 and not isinstance(res_creds[0], dict):
                raise ValueError('res_creds is not a list of credentials')
        else: 
            raise ValueError("No resident credentials found")
        # Construct the passkey dictionary
        passkey = {
            'key': key.get_private(),
            'x5c': certificate,
            'res.creds': res_creds,
            'pin.hash': pin_hash  # Store the full pin hash for future use
        }
        
        return passkey

    @classmethod
    def _save_passkey(cls, key, x5c, resCreds, pinHash, passkeyFilename, secret=None):
        """
        Write passkey file according to the process described in the comments.
        
        Args:
            key: EC Private Key
            x5c: X509 Certificate (CA)
            resCreds: list of resident credentials ({rp.id:<>,user.id:<>,cred.id:<>})
            pinHash: SHA256 hash of the user's PIN
            passkeyFilename: Path to the passkey file
        """
        # Cache upper hash for loading during pin auth protocol
        if len(pinHash) != 32:
            raise ValueError("pinHash must be 32 bytes long")
        #TODO else if upper hash does not match current file, sync error
        upper_hash = pinHash[16:]
        platform_key = cls.__get_platform_kp(secret).get_private()
        header = cls.ec_encrypt(upper_hash, platform_key)
        pkcs12_bytes = cls.create_pcks12_bytes(
            key,
            x5c,
            b"Pirate Passkey Secret Stash",
            base64.b64encode(pinHash),  # Use full pin hash as secret
        )
        
        # Get PKCS12 bytes length
        pkcs12_len = len(pkcs12_bytes)
        pkcs12_len_bytes = pkcs12_len.to_bytes(4, 'little')
        # Encrypt resident credentials with key
        cbor_res_creds = cbor.dumps(resCreds)
        enc_res_creds = cls.ec_encrypt(cbor_res_creds, key)

        # Write file
        body = pkcs12_len_bytes + pkcs12_bytes + enc_res_creds

        if not passkeyFilename.endswith('.passkey'):
            passkeyFilename += '.passkey'
        passkey_path = os.path.join(os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido')), passkeyFilename)
        with open(passkey_path, 'wb') as f:
            f.write(header + body)
            f.close()

    @classmethod
    def __get_platform_kp(cls, secret=None, filename='platform.key'):
        # Get platform key to manage cached pin hashes
        platform_key_path = os.path.join(os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido')), filename)
        with open(platform_key_path, 'rb') as key_file:
            platform_key_pem = key_file.read()
            return KeyPair.load_key_pair(platform_key_pem, secret)

    @classmethod
    def create_platform_key(cls, secret=None, filename='platform.key'):
        plat_key = KeyPair.generate_ecdsa()
        platform_key_path = os.path.join(os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido')), filename)
        with open(platform_key_path, 'wb') as key_file:
            key_file.write(plat_key.get_private_bytes(secret=secret))
        return KeyPair(plat_key, plat_key.get_public())

    @classmethod
    def get_pin_hash(cls, pin, alg=hashes.SHA256()):
        digest = hashes.Hash(alg)
        if not isinstance(pin, bytes):
             pin = pin.encode()
        digest.update(pin)
        return digest.finalize()


    @classmethod
    def create_pcks12_bytes(cls, key, cert, name, secret):
        """
        return pcks12 file bytes for given key, certificate with provided secret.

        Args:
            key: The private key to be serialized.
            cert: The certificate to be serialized.
            name: The name (alias) of the certificate.
            secret: The secret to be used for encryption.

        Returns: The bytes of the serialized pcks12 file.
        """
        return pkcs12.serialize_key_and_certificates(
                name, key, cert, None, serialization.BestAvailableEncryption(secret))


    @classmethod
    def load_key_and_cert(cls, source, password=None):
        """
        Load a private key and X.509 certificate from a file or bytes.
        
        Args:
            source: bytes containing the key and certificate.
            password: Optional password for encrypted keys

        Returns:
            Tuple(KeyPair, x509.Certificate): The loaded key pair and certificate
        """
        
        # Load as PKCS12 first, no additional certs
        private_key, certificate, additional = pkcs12.load_key_and_certificates(source, password)
        if not private_key:
            raise ValueError("Failed to load PKCS12 key")
        if not certificate:
            raise ValueError("Failed to load PKCS12 certificate")
        key_pair = KeyPair(private_key, private_key.public_key())
        return key_pair, certificate
            

    @classmethod
    def ec_encrypt(cls, plaintext, key):
        if not isinstance(key, ec.EllipticCurvePrivateKey):
            raise ValueError("Key must be an EllipticCurvePrivateKey")
        iv = secrets.token_bytes(16)
        anon_kp = KeyPair.generate_ecdsa()
        shared_raw = anon_kp.get_private().exchange(ec.ECDH(), key.public_key())
        
        # Hash the shared secret with SHA-256 to match Java's implementation
        digest = hashes.Hash(hashes.SHA256())
        digest.update(shared_raw)
        shared = digest.finalize()
        
        encryptor = Cipher(algorithms.AES256(shared),
                                                  modes.GCM(iv)).encryptor()
        ciphertext = encryptor.update(plaintext) + encryptor.finalize()
        anon_pub = anon_kp.get_public_bytes()
        annon_pub_bytes = len(anon_pub).to_bytes(4, 'big') + anon_pub
        return annon_pub_bytes + iv + encryptor.tag + ciphertext
    
    @classmethod
    def ec_decrypt(cls, encrypted, key):
        if not isinstance(key, ec.EllipticCurvePrivateKey):
            raise ValueError("Key must be an EllipticCurvePrivateKey")
        pub_bytes_len = int.from_bytes(encrypted[:4], 'big')
        pub_bytes = encrypted[4:pub_bytes_len + 4]
        pubkey = serialization.load_pem_public_key(pub_bytes)
        if not isinstance(pubkey, ec.EllipticCurvePublicKey):
            raise ValueError("Public key must be an EllipticCurvePublicKey")
        ciphertext = encrypted[pub_bytes_len + 4:]
        iv = ciphertext[:16]
        tag = ciphertext[16:32]
        shared_raw = key.exchange(ec.ECDH(), pubkey)
        
        # Hash the shared secret with SHA-256 to match Java's implementation
        digest = hashes.Hash(hashes.SHA256())
        digest.update(shared_raw)
        shared = digest.finalize()
        
        decryptor = Cipher(algorithms.AES256(shared),
                                     modes.GCM(iv, tag=tag)).decryptor()
        return decryptor.update(ciphertext[32:]) + decryptor.finalize()


class KeyPair(object):  

    def __init__(self, privateKey, publicKey):
        object.__init__(self)
        self.private = privateKey
        self.public = publicKey

    @classmethod
    def generate_rsa(cls, e=65537, key_size=2048, backend=default_backend()):
        privateKey = rsa.generate_private_key(e, key_size, backend)
        publicKey = privateKey.public_key()
        return cls(privateKey, publicKey)

    @classmethod
    def generate_ecdsa(cls, curve=ec.SECP256R1(), backend=default_backend()):
        privateKey = ec.generate_private_key(curve, backend)
        publicKey = privateKey.public_key()
        return cls(privateKey, publicKey)

    @classmethod
    def generate_ed25519(cls):
        privateKey = ed25519.Ed25519PrivateKey.generate()
        publicKey = privateKey.public_key()
        return cls(privateKey, publicKey)

    @classmethod
    def generate_mldsa(cls, alg="ML-DSA-44", seed=None):
        from oqs.oqs import Signature
        privateKey: Signature = Signature(alg, secret_key=seed)
        publicKey: bytes = privateKey.generate_keypair()
        return cls(privateKey, publicKey)

    @classmethod
    def load_key_pair(cls, pk, password=None):
        privateKey = serialization.load_pem_private_key(pk, password=password, backend=default_backend())
        publicKey = privateKey.public_key()
        return cls(privateKey, publicKey)

    def set_key(self, privateKey):
        self.private = privateKey
        self.public = privateKey.get_public()

    def get_public(self):
        return self.public

    def get_private(self):
        return self.private

    def get_public_bytes(self):
        return self.public.public_bytes(encoding=serialization.Encoding.PEM,
                                        format=serialization.PublicFormat.SubjectPublicKeyInfo)

    def get_private_bytes(self, secret=None):
        return self.private.private_bytes(encoding=serialization.Encoding.PEM,
                                          format=serialization.PrivateFormat.PKCS8,
                                          encryption_algorithm=serialization.BestAvailableEncryption(secret) if secret 
                                                                else serialization.NoEncryption())



# Copyrite IBM 2022, 2025
# IBM Confidential
# Assisted by watsonx Code Assistant

# CTAP2HID reads bytes from endpoint 1,building the frames into a CTAPMsg. A CTAPMsg is used to initalize a 
# FIDO2Transaction thread which will reply on endpoint 2 when the transaction is complete.
#
# Authenticator uses some environmental properties to read PKI from the filesystem.
#       FIDO_AUTHENTICATOR_PKCS12 :: pkcs12 which contains the CA key/x509 and any generated x509 by the 
#                                    Fido2Authenticator.
#       FIDO_AUTHENTICATOR_PKCS12_SECRET :: password to decrypt the pkcs12 bag.
# Any credentials crated using the CA pkcs12 can be reconstructed during `navigator.credential.get` requests.
#       FIDO2_GENERATE_CA :: boolean to indicate if an error should be thrown if the PKCS12 file does not exist or
#                            cannot be read / converted to requried PKI. If this value is true the CA key and x509
#                            are generated on the first request to create a credential.
#

import base64, datetime, multiprocessing, os, sys, random, queue, threading, time, secrets, traceback
import cbor2 as cbor
from enum import Enum, IntEnum
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import hashes
from cryptography.fernet import Fernet
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from usb_ip import BaseStructure, USBDevice, InterfaceDescriptor, DeviceConfigurations, EndPoint, USBContainer, \
                    USBIPCMDSubmit, bcolors, dump_bytes, colour_print
from key_pair import KeyPair, KeyUtils
from authenticator import Fido2Authenticator
from cert_utils import CertUtils


# Classes to create a USB device object, connect it to the system usb daemon and begin tx/rx
class CTAP2HIDClass(BaseStructure):
    _fields_ = [
        ('bLength', 'B', 9),
        ('bDescriptorType', 'B', 0x21),  # HID
        ('bcdHID', 'H'),
        ('bCountryCode', 'B'),
        ('bNumDescriptors', 'B'),
        ('bDescriptprType2', 'B'),
        ('wDescriptionLength', 'H'),
    ]


hid_class = CTAP2HIDClass(bcdHID=0x0100,  # HID version number
                     bCountryCode=0x0,
                     bNumDescriptors=0x1,
                     bDescriptprType2=0x22,  # Report
                     wDescriptionLength=0x3F00)  # Little endian


interface_d = InterfaceDescriptor(bAlternateSetting=0,
                                  bNumEndpoints=2,
                                  bInterfaceClass=3,  # class HID
                                  bInterfaceSubClass=0, # no interface subclass
                                  bInterfaceProtocol=0, # no interface protocol
                                  iInterface=0)

end_point_one = EndPoint(bEndpointAddress=0x04, # HOST OUT / USB IN
                     bmAttributes=0x3, # Interrupt transfer
                     wMaxPacketSize=(64&0x00FF)<<8 | (64&0xFF00),  # 64-byte packet max
                     bInterval=5)  # Poll every 5 millisecond

end_point_two = EndPoint(bEndpointAddress=0x8E, # HOST IN / USB OUT
                     bmAttributes=0x3, # Interrupt transfer
                     wMaxPacketSize=(64&0x00FF)<<8 | (64&0xFF00),  # 64-byte packet max, bussit
                     bInterval=5)  # Poll every 5 millisecond

interface_d.descriptions = [hid_class]  # Supports only one description
interface_d.endpoints = [end_point_two, end_point_one]  # Supports two endpoint
# wTotalLength = len( interface_d.pack() + hid_class.pack() + end_point_one.pack() + end_point_two.pack() + 18)
configuration = DeviceConfigurations(wTotalLength=0x2900,
                                     bNumInterfaces=0x1,
                                     bConfigurationValue=0x1,
                                     iConfiguration=0x0,  # No string
                                     bmAttributes=0x80,  # valid self powered
                                     bMaxPower=50)  # 50 mah current
configuration.interfaces = [interface_d]   # Supports only one interface


class Authenticator(object):
    '''
    Authenticator wraps the Fido2Authenticator with management for a persistent key/ca certificate stored in
    an encrypted file.

    The default location for this is in the ~/.fido2 directory for the user running this module, however this
    can be overwritten by environment properties.

    All keys generated by this authenticator are created so that the credential id is the encrypted key
    used to generate the attestation. This is done to allow this Authenticator to restore any key created with
    its ca key / AES key combination without the need to persist every individual private key used to 
    generate a new credential.
    '''

    _open_keys = {}
    _lock = multiprocessing.Lock()

    _pin_token_kp = None
    _pin_retry = 5

    def __new__(cls):
        cls._lock.acquire()
        if cls._pin_token_kp == None:
            cls._pin_token_kp = KeyPair.generate_ecdsa() #P-256 key, cose key type -25
        cls._lock.release()

    @classmethod
    def _validate_pin(cls, pinHash, cid):
        '''
        Look in the home direcotry of the user for a .fido2 directory. If it exists and contains
        enough data, try and open it.

        If we can open a given file and it contains the cert/key we expect; add it to cls._open_keys with the 
        pinAuthToken as the key.

        If we successfuly open and validata a file, then a generated pinAuthToken, which will
        function as an opaque lookup key, is returned.
        If no file could be opened then this method will return None
        '''
        # Try all .passkey files in the $HOME/.fido2 dir. If we parse out a cert/key then we have a 
        # valid pin and can cache the result and return an auth token.
        aesKey = algorithms.AES128(pinHash)
        baseDir = os.path.realpath(os.path.join(os.environ.get("HOME"), '.fido2'))
        for maybeKeyFile in os.listdir(baseDir):
            if not maybeKeyFile.endswith('.passkey'):
                colour_print(colour=bcolors.WARNING, component='Authenticator_validate_pin', 
                             msg='{} has invalid file type'.format(maybeKeyFile))
                continue
            encBagF = None
            try:
                encBagF = os.path.join(baseDir, maybeKeyFile)
                passkey = KeyUtils._load_passkey(pinHash, encBagF)
                colour_print(colour=bcolors.OKPINK, component='Authenticator_validate_pin', 
                             msg='Pin decrypted a .passkey file')
                ca = CertUtils.load_der_certificate(passkey.get('x5c'))
                kp = KeyUtils.load_der_key(passkey.get('key'))
                seed = passkey.get('seed')
                cls._pin_retry = 5 #Reset the counter
                pinAuthToken = secrets.token_bytes(32)
                cls._open_keys[cid] = {
                            'fKey': Fernet(seed),
                            'pem': ca,
                            'kp': kp,
                            'file': encBagF,
                            'ph': pinHash
                        }
                return pinAuthToken
            except Exception as e:
                colour_print(colour=bcolors.WARNING, component='Authenticator_validate_pin', 
                             msg='Something bad happened:\n{}'.format(e))
                continue
        colour_print(colour=bcolors.FAIL, component='Authenticator_validate_pin', 
                     msg='No valid pin found!')
        return None

    @classmethod
    def get_pin_cose_key(cls, pin_req, cid):
        if cls._pin_token_kp == None:
            cls.__new__(cls)
        return {1: KeyUtils.get_cose_key(cls._pin_token_kp.get_public(), hashes.SHA256(), ecdh=True)}

    @classmethod
    def get_pin_retries(cls, pin_req, cid):
        cls._pin_retry -= 1
        return {3: cls._pin_retry}

    @classmethod
    def decapsulate(cls, ecCoseKey):
        cose_type_to_curve_map = { #These are kind of made up, as per 
                    #https://fidoalliance.org/specs/fido-v2.1-ps-20210615/fido-client-to-authenticator-protocol-v2.1-ps-errata-20220621.html#pinProto1
                    -25: ec.SECP256R1,
                    -26: ec.SECP521R1
                }
        ec_pub_numbs = ec.EllipticCurvePublicNumbers(KeyUtils._bytes_to_long(ecCoseKey[-2]),
                            KeyUtils._bytes_to_long(ecCoseKey[-3]), 
                            cose_type_to_curve_map[ecCoseKey[3]]())
        pubkey = ec_pub_numbs.public_key()
        shared_point = cls._pin_token_kp.get_private().exchange(ec.ECDH(), pubkey)
        hasher = hashes.Hash(hashes.SHA256());
        hasher.update(shared_point)
        return hasher.finalize()

    @classmethod
    def get_pin_token(cls, pin_req, cid):
        #https://fidoalliance.org/specs/fido-v2.1-ps-20210615/fido-client-to-authenticator-protocol-v2.1-ps-errata-20220621.html#getPinToken
        platform_cose_key = pin_req[3]
        pin_hash_enc = pin_req[6]
        colour_print(colour=bcolors.OKPINK, component='Authenticator.get_pin_token', 
                     msg='plat cose key: {}; pinHashEnc: {}'.format(platform_cose_key, pin_hash_enc))
        sharedSecret = cls.decapsulate(platform_cose_key)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.get_pin_token', 
                     msg='shared secret: {};'.format(sharedSecret))
        decryptor = Cipher(algorithms.AES256(sharedSecret), modes.CBC(bytes([0] * 16))).decryptor()
        pin = decryptor.update(pin_hash_enc) + decryptor.finalize()
        pinAuthToken = cls._validate_pin(pin, cid)
        if pinAuthToken != None:
            return {2: pinAuthToken}
        return None


    @classmethod
    def attestation_out(cls, clientDataHash, rp, user, pkCredsParams, excludeList, exts, cid):
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='open keys: {}'.format(cls._open_keys))
        if not cid in cls._open_keys.keys():
            return None, None
        ca = cls._open_keys.get(cid)
        kp = KeyPair.generate_ecdsa()
        authenticator = Fido2Authenticator(keyPair=kp, caKeyPair=ca.get('kp'), 
                                            caCert=ca.get('pem'), fKey=ca.get('fKey'))
        credId = authenticator.get_credential_id(kp)
        pk = {'rp': rp} # authenticator.py expects this strucutre
        authData = authenticator.build_authenticator_data(pk, 'packed', kp, True, up=True, be=False, bs=False)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='credId: {}; toSign: {}'.format(credId, base64.b64encode(bytes([*authData, *clientDataHash ])).decode()))
        attStmt = authenticator.build_packed_attestation_statement('packed', 
                                                    clientDataHash, authData, None, kp)
        colour_print(colour=bcolors.OKPINK, component='Authenticator.attestation_out', 
                     msg='attStmt: {}'.format(attStmt))
        #Since we set UV we can make these resident keys
        KeyUtils.update_passkey({rp['id']: credId, 'user': user['id']}, ca['ph'], ca['file'])
        return authData, attStmt


    @classmethod
    def assertion_out(cls, rpId, clientDataHash, allowedList, exts, cid):
        if not cid in cls._open_keys.keys():
            return None, None, None, None
        ca = cls._open_keys.get(cid)
        resCreds = KeyUtils._load_passkey(ca['ph'], ca['file']).get('res_creds')
        if resCreds != None:
            colour_print(colour=bcolors.OKPINK, component='FIDO2Authenticator.assertion_outputs',
                         msg='passkey has resident credentials, adding them to allowed list')
            for rc in resCreds:
                if rc.get(rpId) != None:
                    allowedList += [{'id': base64.urlsafe_b64decode(rc[rpId]), 
                                     'user': rc.get('user')}]
        for cred in allowedList:
            try:
                ca_pem = ca.get('pem')
                ca_kp = ca.get('kp')
                fKey = ca.get('fKey')
                #Create the authenticator, creating the key pair will fail if we don'w own the credId
                b64CredId = base64.urlsafe_b64encode(cred.get('id'))
                kp = Fido2Authenticator._get_key_pair_from_credential_id(b64CredId, fKey)
                colour_print(colour=bcolors.OKPINK, component='FIDO2Authenticator.assertion_outputs',
                             msg='We have a usable key, sign the challenge')
                _authenticator = Fido2Authenticator(keyPair=kp, credId=b64CredId, aaguid=[b'\x00'*16], 
                                                    caKeyPair=ca_kp, caCert=ca_pem, fKey=fKey)
                credential = {
                        "id": _authenticator.cib, #Should be set by __init__()
                        "type" : "public-key"
                    }
                #Generate the assertion response data
                authData = _authenticator.build_authenticator_data({'rpId': rpId}, 'packed', kp, True, 
                                                                        up=True, be=False, bs=False)
                #Sign it, authenticator method b64 encodeds the ba so we decode it right away
                sig = _authenticator.assertion_signature(authData, clientDataHash, kp)
                userHandle = cred.get("user")
                return credential, authData, sig, userHandle
            except Exception as e:
                colour_print(colour=bcolors.FAIL, component='FIDO2Authenticator.assertion_outputs',
                             msg='Could not retrieve key pair from credential id {}'.format(cred))
                continue
        return None, None, None, None

class CBORCommand(object):

    class CommandByte(Enum):
        MAKE_CREDENTIAL = 0x1
        GET_NEXT_ASSERTION = 0x2
        GET_INFO = 0x4
        CLIENT_PIN = 0x6
        RESET = 0x7
        CREDENTIAL_MANAGEMENT = 0x9
        AUTHENTICATOR_SELECTION = 0xB
        AUTHENTICATOR_CONFIG = 0xD

        def __repr__(self):
            #return f'{self.__class__.__name__}.{self.name}'
            return self.value

    class CBORStatusCode(IntEnum):
        CTAP2_OK = 0x0
        CTAP1_ERR_INVALID_COMMAND = 0x01
        CTAP2_ERR_INVALID_CBOR = 0x12
        CTAP2_ERR_PIN_INVALID = 0x31
        CTAP2_ERR_PIN_AUTH_INVALID = 0x33
        CTAP2_ERR_PUAT_REQUIRED = 0x36


    cid = 0xFFFFFFFF
    request = []
    response = []
    response_segment = 0
    response_ready = False
    length = 0
    request_segment = 0
    cmd = None
    ctaphid_cmd = 0
    bcnt = 0

    def __init__(self, cid, ba, skip_init=False):
        self.cid = cid
        if ba == None and skip_init == True:
            return #Create an empty command as we will directly set the response buffer later with the assigned CID.
        if len(ba) <= 1:
            colour_print(colour=bcolors.OKYELLOW, component='CBORCommand.__init__', 
                    msg="Byte Array must be at least one byte long")
        # Length of the incoming CBOR message (total).
        self.length = int.from_bytes(ba[0:2], 'big')
        # Request buffer. This stores the incoming CBOR message and grows until all segments have been recieved
        self.request = ba[3:]
        # Track then number of response segments transmitted, the number transmitted in the continue sequence packet
        # should always be one less than this number
        self.response_segment = 0
        # Track the number of request segmetns recieved
        self.request_segment = 0
        # Response buffer. This stores the outgoing response to the recieved CBOR message and shrinks until the entire
        # response has been transmitted
        self.response = []
        # Command recieved in CTAPHID frame, this is likely 0x90 (CBOR_MSG) but might be different
        self.ctaphid_cmd = 0
        # Authenticator API command byte recieved in initial packet
        self.cmd = self.CommandByte(int.from_bytes(ba[2:3], 'big'))
        #Length of the payload bytes
        self.bcnt = 0
        # Signal that the response buffer is ready to be sent back to the client
        self.response_ready = False
        colour_print(colour=bcolors.OKPURPLE, component='CBORCommand.__init__', 
                msg="command {}; length {}; self.request[{}]".format(self.cmd, self.length, len(self.request)))
        if self.length >= len(self.request):
            colour_print(colour=bcolors.OKPURPLE, component='CBORCommand.__init__', 
                    msg="request is segmented, wait for the whole message")
        else: #We have the whole message
            self.response = self.unpack()
            dump_bytes(self.response, colour=bcolors.OKPINK, 
                       component='CBORCommand.__init__', msg='CTAP response')

    def append_segment(self, seg_buf):
        self.request += seg_buf
        if len(self.request) >= self.length:
            self.response = self.unpack()
            dump_bytes(self.response, colour=bcolors.OKPINK, 
                       component='CBORCommand.append_segment', msg='CTAP response')
        else:
            self.request_segment += 1


    #Return CBOR response if entire command has been received or None if still 
    #waiting for segments
    def unpack(self):
        return {
            self.CommandByte.MAKE_CREDENTIAL: self._make_cred,
            self.CommandByte.GET_NEXT_ASSERTION: self._get_assertion,
            self.CommandByte.GET_INFO: self._get_info,
            self.CommandByte.CLIENT_PIN: self._client_pin
            }.get(self.cmd)(bytes(self.request))

    def get_rsp_seg(self, num_bytes):
        if not isinstance(num_bytes, int):
            raise RuntimeError("panic!")
        self.response_segment += 1
        #sequence is offset by two to account for init pkt and zero index start for continue sequence
        seg_num = max(self.response_segment - 2, 0)
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', msg='self.response_segment = {}'.format(self.response_segment))
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', msg='self.response_segment - 2 = {}'.format(self.response_segment - 2))
        colour_print(colour=bcolors.WARNING, component='CBORCommand.get_rsp_seg', msg='segment number = {}'.format(seg_num))
        seg = self.response
        if num_bytes >= len(self.response):
            self.response = []
        else:
            seg = self.response[:num_bytes]
            self.response = self.response[num_bytes:]
        return seg, seg_num

    def _client_pin(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorClientPIN
        pin_sub_cmds = { 
                      1: Authenticator.get_pin_retries,
                      2: Authenticator.get_pin_cose_key,
            #SET_PIN = 0x3
            #CHANGE_PIN = 0x4
                      5: Authenticator.get_pin_token
                }
        req_data = cbor.loads(ba)
        colour_print(colour=bcolors.OKGREEN, component='CBORCommand._client_pin',
                     msg='Packet request: {}'.format(req_data))
        sub_cmd = req_data[2]
        colour_print(colour=bcolors.OKGREEN, component='CBORCommand._client_pin',
                     msg='pin sub_cmd: {}'.format(sub_cmd))
        rsp = pin_sub_cmds[sub_cmd](req_data, self.cid)
        result = (self.CBORStatusCode.CTAP2_ERR_PIN_INVALID).to_bytes()
        if rsp != None:
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes() + cbor.dumps(rsp)
        self.bcnt = len(result)
        self.response_ready = True
        return result

    # authenticatorGetInfo takes no inputs so return immediately
    def _get_info(self, ba):
        result = {
            0x01: ["U2F_V2", "FIDO_2_0"],
            0x02: ['hmac-secret'],
            #0x03: b"\x01\x02\x03\x04" * 4,
            0x03: b"\x00" * 16,
            0x04: {'rk': True, 'up': True, 'plat': False, 'clientPin': True},
            0x05: 1200,
            0x06: [1]
        }
        result = (self.CBORStatusCode.CTAP2_OK).to_bytes() + cbor.dumps(result)
        self.bcnt = len(result)
        self.response_ready = True
        return result

    def _make_cred(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorMakeCredential
        req = cbor.loads(ba)
        colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                     msg='CBOR request {}'.format(req))
        for prop in [(0x01, 'clientDataHash'), (0x02, 'rp'), (0x03, 'user'), (0x04, 'pubkeyCredParams')]:
            if not prop[0] in req.keys():
                colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                             msg='{} missing from request:\n{}'.format(prop[1], cbor.dumps(req)))
                raise Exception("Missing required property %s" % prop[1])
        authData, attStmt = Authenticator.attestation_out(req.get(0x01), req.get(0x02), req.get(0x03),
                                            req.get(0x04), req.get(0x05), req.get(0x06), self.cid)
        result = (self.CBORStatusCode.CTAP2_ERR_PUAT_REQUIRED).to_bytes()
        if authData and attStmt:
            rsp = {
                0x01: 'packed', #fmt
                0x02: authData,
                0x03: attStmt
            }
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes() + cbor.dumps(rsp)
        self.bcnt = len(result)
        self.response_ready = True
        return result


    def _get_assertion(self, ba):
        # https://fidoalliance.org/specs/fido-v2.2-rd-20230321/fido-client-to-authenticator-protocol-v2.2-rd-20230321.html#authenticatorGetAssertion
        req = cbor.loads(ba)
        colour_print(colour=bcolors.FAIL, component='CBORCommand._get_assertion',
                     msg='CBOR request {}'.format(req))
        for prop in [(0x01, 'rpId'), (0x02, 'clientDataHash')]:
            if not prop[0] in req:
                colour_print(colour=bcolors.FAIL, component='CBORCommand._make_cred',
                             msg='{} missing from request:\n{}'.format(prop[1], cbor.dumps(req)))
                raise Exception("Missing required property %s" % prop[1])
        credential, authData, signature, userHandle = Authenticator.assertion_out(req.get(0x01), 
                                                req.get(0x02), req.get(0x03, []), req.get(0x04, {}), self.cid)
        result = (self.CBORStatusCode.CTAP2_ERR_PUAT_REQUIRED).to_bytes()
        if credential and authData and signature:
            rsp = {
                    0x01: credential,
                    0x02: authData,
                    0x03: signature
            }
            if userHandle:
                rsp[0x04] = {'id': userHandle}
            result = (self.CBORStatusCode.CTAP2_OK).to_bytes() + cbor.dumps(rsp)
        self.bcnt = len(result)
        self.response_ready = True
        return result

# Send this for every new message frame. Response contains
class CTAPHIDInitPkt(BaseStructure):
    _fields_ = [
        ('cid', 'I'),
        ('cmd', 'B'),
        ('bcnt', 'H'),
    ]

    def __init__(self, **kwargs):
        if 'data' in kwargs:
            index = None
            for i, field in enumerate(self._fields_):
                if field[0] == 'data':
                    index = i
                    break
            if index == None:
                colour_print(colour=bcolors.OKGREEN, component='CTAPHIDInitPkt.__init__', msg='setting data field')
                self._fields_ += [('data', '%ds' % len(kwargs['data']))]
            else:
                colour_print(colour=bcolors.OKGREEN, component='CTAPHIDInitPkt.__init__', 
                             msg='data already exists as a field, updating it')
                self._fields_[index] = ('data', '%ds' % len(kwargs['data']))
            print(self._fields_)
        super().__init__(**kwargs)

class CTAPHIDSeqPkt(BaseStructure):

    _fields_ = [
        ('cid', 'I'),
        ('seq', 'B'),
    ]

    def __init__(self, **kwargs):
        print(kwargs)
        if 'data' in kwargs:
            index = None
            for i, field in enumerate(self._fields_):
                if field[0] == 'data':
                    index = i
                    break
            if index == None:
                print("setting data field")
                colour_print(colour=bcolors.OKPINK, component='CTAPHIDSeqPkt.__init__', msg='setting data field')
                self._fields_ += [('data', '%ds' % len(kwargs['data']))]
            else:
                colour_print(colour=bcolors.OKPINK, component='CTAPHIDSeqPkt.__init__', 
                             msg='data already exists as a field, updating it')
                self._fields_[index] = ('data', '%ds' % len(kwargs['data']))
            print(self._fields_)
        super(CTAPHIDSeqPkt, self).__init__(**kwargs)



class CTAP2HIDevice(USBDevice):
    speed = 2
    vendorID = 0x3713
    productID = 0x3713
    bcdDevice = 0x0
    bcdUSB = 0x0
    bNumConfigurations = 0x1
    bNumInterfaces = 0x1
    bConfigurationValue = 0x1
    bDeviceClass = 0x0
    bDeviceSubClass = 0x0
    bDeviceProtocol = 0x0
    bNumConfigurations = 1
    bConfigurationValue = 1
    bNumInterfaces = 1
    configurations = [configuration]  # Supports only one configuration

    #This will contain the current set of channel id's and associated state
    cids = {}
    #This is a list of pending message frames with a worker thread which will respond within 50ms if 
    # we don't generate a response in time
    pending = []

    def __init__(self):
        USBDevice.__init__(self)
        self.start_time = datetime.datetime.now()

    def unlink(self, usb_req):
        index = None
        for i, worker in enumerate(self.pending):
            if worker.req.seqnum == usb_req.unlink_seqnum:
                index = i
        if index != None:
            del self.pending[index]
            colour_print(component='CTAP2HIDevice.unlink', msg='unlinked request {}'.format(usb_req.unlink_seqnum))
            return True
        return False

    def generate_fido2_report(self):
        arr = [0x06, 0xd0, 0xf1,        # USAGE_PAGE (FIDO Alliance)
                0x09, 0x01,             # USAGE (CTAPHID)
                0xa1, 0x01,             # HID_Collection(HID_Application)
                0x09, 0x20,             # USAGE (Input Report Data)
                0x15, 0x00,             # LOGICAL_MINIMUM (0)
                0x26, 0xff, 0x00,       # LOGICAL_MAXIMUM (255)
                0x75, 0x08,             # REPORT_SIZE (8)
                0x95, 0x40,             # REPORT_COUNT (64)
                0x81, 0x02,             # INPUT (Data,Var,Abs)
                0x09, 0x21,             # USAGE(Output Report Data)
                0x15, 0x00,             # LOGICAL_MINIMUM (0)
                0x26, 0xff, 0x00,       # LOGICAL_MAXIMUM (255)
                0x75, 0x08,             # REPORT_SIZE (8)
                0x95, 0x40,             # REPORT_COUNT (64)
                0x91, 0x02,             # OUTPUT (Data,Var,Abs)
            0xc0]                   # End Collection
        return bytes(arr) 

    def _bytes_to_str(self, b):
        return ''.join("%02X" % x for x in b)


    class KeepAliveWorker(object):

        stop = False
        my_thread = None
        req = None
        hid = None
        start = None

        def __init__(self, req, hid):
            self.req = req
            self.hid = hid

        def reply_with_keepalive(self):
            start = round(time.time() * 1000)
            while self.stop == False:
                now = round(time.time() * 1000)
                wait_time = now - start
                if(now >= start + 450) and self.stop == False: #If we get to 45ms, send heartbeat
                    colour_print(colour=bcolors.FAIL, component='KeepAliveWorker.reply_with_keepalive',
                                 msg='Thread reached timeout of {} ms before response buffer was recieved . . . sending heartbeat response'.format(wait_time))
                    #We need to reply with keep alive after 50ms this gives us some tolerance
                    #Our status is always still processing
                    rsp = b'\x3B\x01\x01'
                    self.hid.send_usb_req(rsp, len(rsp), start_frame=0xFFFFFFFF, packets=0, ep=0, direction=0, 
                                       seqnum=self.req.seqnum)
                    self.stop = True
                    for idx, txn in enumerate(self.hid.pending): #If we sent a keep alive remove the transaction from the
                        # list of pending transactions
                        if txn.req is not None and txn.req.seqnum is not None and txn.req.seqnum == self.req.seqnum:
                            del self.hid.pending[idx]
                    return
                time.sleep(5/1000) # sleep for 5 ms


    def send_response_segment(self, cid, cbor_cmd):
        if len(self.pending) == 0:
            colour_print(colour=bcolors.FAIL, component='send_response_segment', msg='No pending transactions to use :(')
            return
        #Take control of the next pending request, if this fails then we will just have to wait for the
        # next frame to appear.
        self.pending[0].stop = True
        if self.pending[0].my_thread is not None and self.pending[0].my_thread.is_alive() == True:
            self.pending[0].my_thread.join(1/1000)
        if self.pending[0].my_thread.is_alive() == True:
            colour_print(colour=bcolors.FAIL, component='send_response_segment', msg='Could not kill keepalive thread')
            return
        rsp_data = None
        if cbor_cmd.response_segment == 0: #We send the init pkt
            data, _ = cbor_cmd.get_rsp_seg(57)
            rsp_data = CTAPHIDInitPkt(cid=int.from_bytes(cid, 'big'), 
                                    cmd=cbor_cmd.ctaphid_cmd,
                                    bcnt=cbor_cmd.bcnt,
                                    data=data).pack()
        else: #We send the continue sequence pkt
            data, seq_num = cbor_cmd.get_rsp_seg(59)
            colour_print(colour=bcolors.WARNING, component='send_response_segment', msg='Sequence number {}'.format(seq_num))
            rsp_data = CTAPHIDSeqPkt(cid=int.from_bytes(cid, 'big'),
                                     seq=seq_num,
                                     data=data).pack()
        colour_print(colour=bcolors.WARNING, component='send_response_segment', msg='pad with {} 0 bytes'.format(64 - len(rsp_data)))
        rsp_data += b'\00' * (64 - len(rsp_data)) # pad the 64 byte frame with 0x00 if required
        '''
        rsp_data = CTAPHIDInitPkt(cid=int.from_bytes(cid), cmd=int.from_bytes(cmd), bcnt=17, data=data).pack()
        dump_bytes(rsp, colour=bcolors.OKGREEN, component='USBDevice.ctaphid_init', msg='Packed response')
        rsp_data = cid
        rsp_data += int.to_bytes(cbor_cmd.response_segment, 1)
        if len(cbor_cmd.response) > 58:
            rsp_data += cbor_cmd.response[:58]
            cbor_cmd.response = cbor_cmd.response[58:]
        else:
            rsp_data += cbor_cmd.response
            cbor_cmd.response = []
        # Pad out to 64 bytes
        rsp_data += b'\x00' * (64 - len(rsp_data))
        '''
        dump_bytes(rsp_data, colour=bcolors.OKGREEN, component='USBDevice.send_response_segment', 
                   msg='Packed response: ')
        self.send_usb_req(rsp_data, len(rsp_data), start_frame=0xFFFFFFFF, packets=0, 
                          ep=0, direction=0, seqnum=self.pending[0].req.seqnum)
        del self.pending[0]
        #If we have not response buffer left, remove the transaction from our context
        # don't do this for the broadcast channel.
        if len(cbor_cmd.response) == 0 and cid != b'\xff\xff\xff\xff':
            del self.cids[cid]['cborCmd']
        return


    def ctaphid_ping(self, usb_req):
        return

    def ctaphid_msg(self, usb_req):

        class U2FCommand(Enum):
            U2F_VERSION = 0x0
            U2F_REGISTER = 0x1
            U2F_AUTHENTICATE = 0x2
            U2F_VER = 0x03

        #Only supporting extended length encoding, section 3.1.3
        #https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.html
        cid = usb_req.data_frame[0:4]
        if not cid in self.cids:
            colour_print(colour=bcolors.FAIL, component='USBDevice.ctaphid_msg', 
                         msg='Unknown CID {}'.format(cid))
        cmd = usb_req.data_frame[4:5]
        bcnt = usb_req.data_frame[5:7]
        apdu = usb_req.data_frame[7:]
        colour_print(colour=bcolors.OKGREEN, component='USBDevice.ctaphid_msg', 
                    msg='cmd = {}; bcnt = {}; apdu = {}'.format(
                        self._bytes_to_str(cmd), self._bytes_to_str(bcnt), apdu))
        u2f_cla = apdu[:1]
        u2f_ins = apdu[1:2]
        u2f_p1 = apdu[2:3]
        u2f_p2 = apdu[3:4]
        u2f_lc = apdu[4:7]
        u2f_data = apdu[7:]
        colour_print(colour=bcolors.OKGREEN, component='USBDevice.ctaphid_msg', 
                    msg='U2F Raw Message CLA = {}; INS = {}; P1 = {}; P2 = {}; Lc = {}'.format(
                        u2f_cla, u2f_ins, u2f_p1, u2f_p2, u2f_lc))
        if u2f_cla == b'\x00' and u2f_ins == b'\x03':
            #U2F_VERSION request, send the expected response
            cborCmd = CBORCommand(cid, None, skip_init=True)
            cborCmd.ctaphid_cmd = int.from_bytes(cmd, 'big')
            cborCmd.bcnt = 6
            cborCmd.response = b'U2F_V2'
            self.cids[cid]['cborCmd'] = cborCmd
            self.send_response_segment(cid, self.cids[cid]['cborCmd'])

    def ctaphid_init(self, usb_req):
        cid = usb_req.data_frame[0:4]
        cmd = usb_req.data_frame[4:5]
        bcnt = usb_req.data_frame[5:7]
        nonce = usb_req.data_frame[7:15]
        assignedCID = bytes([0, random.randint(0, 255), 0, random.randint(0, 255)])
        colour_print(colour=bcolors.OKGREEN, component='USBDevice.ctaphid_init', 
                    msg='Assigning a new CID to {}'.format(self._bytes_to_str(assignedCID)))
        data = nonce + assignedCID
        # protocol == 2; major version == 5; minor version = 2; build version = 5
        # data += int.to_bytes(2) + int.to_bytes(5) + int.to_bytes(1) + int.to_bytes(2) + int.to_bytes(5)
        for i in [2, 5, 1, 2, 0x04 | 0x08]:
            data += int.to_bytes(i)
        dump_bytes(data, colour=bcolors.OKGREEN, component='USBDevice.ctaphid_init', msg='Response data')
        data += b'\00' * (57 - len(data)) # 64 - 4 (CID) - 1 (cmd) - 2 (bcnt) - len of response
        #rsp = CTAPHIDInitPkt(cid=int.from_bytes(cid), cmd=int.from_bytes(cmd), bcnt=17, data=data).pack()
        #dump_bytes(rsp, colour=bcolors.OKGREEN, component='USBDevice.ctaphid_init', msg='Packed response')
        self.cids[assignedCID] = {'cborCmd': CBORCommand(cid, None, skip_init=True) }
        self.cids[assignedCID]['cborCmd'].response = data
        self.cids[assignedCID]['cborCmd'].ctaphid_cmd = int.from_bytes(cmd, 'big')
        self.cids[assignedCID]['cborCmd'].bcnt = 17
        #return self.send_usb_req(usb_req, rsp, 64, ep=0x81, start_frame=int.from_bytes(cid))
        #self.send_usb_req(usb_req, b'', 0, ep=usb_req.ep)
        #self.send_usb_req(rsp, len(rsp), start_frame=0xFFFFFFFF, packets=0, 
        #                  ep=0, direction=0, seqnum=self.pending[0].seqnum)
        #del self.pending[0]
        self.send_response_segment(cid, self.cids[assignedCID]['cborCmd'])

    def ctaphid_cbor(self, usb_req):
        cid = usb_req.data_frame[0:4]
        colour_print(colour=bcolors.OKGREEN, component='USBDevice.ctaphid_cbor', 
                    msg='CBOR message recieved on channel {}'.format(self._bytes_to_str(cid)))
        cmd = usb_req.data_frame[4:5]
        bcnt = usb_req.data_frame[5:7]
        ctap_cmd = usb_req.data_frame[7:8]
        print(int.from_bytes(bcnt, 'big') - 1)
        cbor_data = usb_req.data_frame[8: 7 + int.from_bytes(bcnt, 'big')]
        colour_print(colour=bcolors.OKGREEN, component='USBDevice.ctaphid_cbor', 
                     msg='CBOR msg frame cmd: {}; bcnt: {}'.format(self._bytes_to_str(ctap_cmd),
                                                                   self._bytes_to_str(bcnt)))
        dump_bytes(cbor_data, colour=bcolors.OKGREEN, component='USBDevice.ctaphid_cbor', 
                    msg='CBOR encoded bytes: ')
        cbor_cmd = CBORCommand(cid, usb_req.data_frame[5:])
        cbor_cmd.ctaphid_cmd = int.from_bytes(cmd, 'big')
        self.cids[cid]['cborCmd'] = cbor_cmd
        if cbor_cmd.response_ready == True: #We can respond immediatly
            dump_bytes(self.cids[cid]['cborCmd'].response, colour=bcolors.OKGREEN, 
                       component='USBDevice.ctaphid_cbor', msg='CBOR response: ')
            return self.send_response_segment(cid, self.cids[cid]['cborCmd'])
        else:
            colour_print(colour=bcolors.OKYELLOW, component='USBDevice.ctaphid_cbor', 
                         msg="Waiting for rest of command to arrive . . .")
            return

    def ctaphid_cancel(self, usb_req):
        return

    def ctaphid_keepalive(self, usb_req):
        return

    def ctaphid_error(self, usb_req):
        return

    def ctaphid_unknown(self, usb_req):
        colour_print(colour=bcolors.FAIL, component='USBDevice.ctaphid_unknown', msg='Unkown request recieved')
        self.send_usb_req(b'', 0, ep=usb_req.ep, seqnum=usb_req.seqnum)

    def _handle_outgoing(self, usb_req):
        colour_print(component='USBDevice._handle_outgoing', msg='Adding request to pending')
        # Create a new thread so we will reply at 50ms if required
        worker = self.KeepAliveWorker(usb_req, self)
        keep_alive_thread = threading.Thread(target=worker.reply_with_keepalive)
        worker.my_thread = keep_alive_thread
        #keep_alive_thread.start()
        self.pending.append( worker )
        for cid, context in self.cids.items():
            if context.get('cborCmd') != None and context['cborCmd'].response_ready == True:
                colour_print(component='USBDevice._handle_outgoing', 
                            msg='Using pending request to send response segment')
                self.send_response_segment(cid, context['cborCmd'])
                return

    def _handle_incoming_cmd(self, cmd, usb_req):
        ctapCmd = int.from_bytes(cmd, 'big') & 0x7F
        colour_print(colour=bcolors.OKGREEN, component='USBDevice._handle_incoming_cmd', 
                    msg='recieved command {}'.format(ctapCmd))
        return {
            1: self.ctaphid_ping,
            3: self.ctaphid_msg,
            6: self.ctaphid_init,
            16: self.ctaphid_cbor,
            17: self.ctaphid_cancel,
            59: self.ctaphid_keepalive,
            63: self.ctaphid_error,
        }.get(ctapCmd, self.ctaphid_unknown)(usb_req)

    def _handle_incoming_sequence(self, cid, usb_req):
        seqNum = int.from_bytes(usb_req.data_frame[4:5], 'big')
        context = self.cids.get(cid)
        if context != None:
            transaction = context.get("cborCmd")
            if transaction != None and seqNum == transaction.request_segment:
                transaction.append_segment(usb_req.data_frame[5:])
                if transaction.response_ready == True:
                    self.send_response_segment(cid, transaction)
                else:
                    colour_print(colour=bcolors.OKPURPLE, component='USBDevice._handle_incoming_sequence', 
                                 msg='Sequence number [{}] not the last expected sequence'.format(seqNum))
            else:
                colour_print(colour=bcolors.FAIL, component='USBDevice._handle_incoming_sequence', 
                             msg='Sequence number [{}] not the next expected sequence [{}]'.format(
                                 seqNum, transaction.request_segment))
        else:
            colour_print(colour=bcolors.FAIL, component='USBDevice._handle_incoming_sequence', 
                         msg='CID not found in device context, don\'t know what to do')

    def _handle_incoming(self, usb_req):
        if len(self.pending) == 0:
            colour_print(colour=bcolors.FAIL, component="USBDevice._handle_incoming", 
                         msg="No pending request to respond with :(")
        else: #We reply to the HOST Out endpoint with an empty frame then process the command
            self.send_usb_req('', 64, ep=0, start_frame=0xFFFFFFFF, seqnum=usb_req.seqnum)
        # We send the response to the command to a request in self.pending once we are ready
        cid = usb_req.data_frame[0:4]
        cmd = usb_req.data_frame[4:5]
        colour_print(colour=bcolors.OKGREEN, component='USBDevice._handle_incoming', 
                    msg='CID: {}; command: {}'.format(self._bytes_to_str(cid), self._bytes_to_str(cmd)))
        if(int.from_bytes(cmd, 'big') & 0x80) > 0:
            colour_print(colour=bcolors.FAIL, component='USBDevice._handle_incoming', 
                        msg='bit 8 set we got a command msg')
            return self._handle_incoming_cmd(cmd, usb_req)
        else:
            colour_print(colour=bcolors.OKPURPLE, component='USBDevice._handle_incoming',
                         msg='Recieved a sequence segment, appending it to the current msg context')
            return self._handle_incoming_sequence(cid, usb_req)

    def handle_data(self, usb_req):
        '''
        if usb_req.start_frame == 0xFFFFFFFF:
            colour_print(component='USBDevice.handle_data', msg='Don\'t know what to do with this request')
            self.send_usb_req(b'', 0, seqnum=usb_req.seqnum)
        '''
        if usb_req.ep == 0xE: #HOST In endpoint, add it to the queue so we can use it when we have data 
            return self._handle_outgoing(usb_req) # or use it to send the current in progress transaction
        else: #We have data, work out what to do with it and if we have a pending request we can respond with
            return self._handle_incoming(usb_req)


    def handle_unknown_control(self, control_req, usb_req):
        handled = False
        if control_req.bmRequestType == 0x81: #Interface request
            if control_req.bRequest == 0x6:  # Get Descriptor
                if control_req.wValue == 0x22:  # send initial report
                    print('[' + bcolors.OKGREEN + 'USBDevice' + bcolors.ENDC + '] Send initial report ')
                    ret=self.generate_fido2_report()
                    self.send_usb_req(ret, len(ret), seqnum=usb_req.seqnum)
                    handled = True
        elif control_req.bmRequestType == 0x21:  # Host Request
            if control_req.bRequest == 0x0a:  # set idle
                print('[' + bcolors.OKGREEN + 'USBDevice' + bcolors.ENDC + '] HID Idle ')
                self.send_usb_req(b'', 0, seqnum=usb_req.seqnum)
                handled = True
        else:
            print('[' + bcolors.FAIL + 'USBDevice' + bcolors.ENDC + '] Unknown control [{}] '.format(
                ', '.join( hex(x) for x in list(usb_req.setup.to_bytes(8, 'big')) )))
            print('[' + bcolors.FAIL + 'USBDevice' + bcolors.ENDC + '] Unknown flags [{}] '.format(
                ', '.join( hex(x) for x in list(usb_req.flags.to_bytes(8, 'big')) )))
            print('[' + bcolors.FAIL + 'USBDevice' + bcolors.ENDC + '] Unknown data [{}] '.format(
                ', '.join( hex(x) for x in list(usb_req.data_frame) )))
            #raise RuntimeError("unknown control")
            self.send_usb_req(b"\x01\x00",2, seqnum=usb_req.seqnum)
            pass
        return handled


'''
To run:
    on SERVER::
    $ python hid_device.py

    on CLIENT::
    sudo modprobe vhci-hcd
    sudo usbip -d list -r 127.0.0.1
    sudo usbip -d attach -r 127.0.0.1 -b 1-1.1
    lsusb -v -d 1337:1337
'''

usb_dev = CTAP2HIDevice()
usb_container = USBContainer()
usb_container.add_usb_device(usb_dev)  # Supports only one device!
usb_container.run()



#!/bin/python
# Copyrite IBM 2022, 2025
# IBM Confidential

import logging, sys, os
try:
    from .passkey_device import CTAP2HIDevice
    from .systray_app import SysTrayApp
except:
    try:
        sys.path.append(os.path.dirname(os.path.abspath(__file__)))
        from passkey_device import CTAP2HIDevice
        from systray_app import SysTrayApp
    except Exception as e:
        logging.debug("Module load error")
        logging.exception(e)
        raise e


if os.environ.get("FIDO_HOME") == None:
    sys.exit(1)
ll = logging.INFO
if "SOFT_FIDO2_DEBUG_LEVEL" in os.environ:
    ll = os.environ.get("SOFT_FIDO2_DEBUG_LEVEL")
logFile = None # > stdout/stderr
if "SOFT_FIDO2_LOG_FILE" in os.environ:
    logFile = os.path.join(
                        os.environ["FIDO_HOME"], os.environ["SOFT_FIDO2_LOG_FILE"])

#logPath = os.path.join(os.environ.get("FIDO_HOME"), 'passkey.log')
logging.basicConfig(level=ll, format='%(message)s', filename=logFile)
#logging.basicConfig(level=logging.DEBUG, format='%(message)s')
logging.info("Starting the EyeBeeKey Passkey UHID Service")
print("Starting the EyeBeeKey Passkey UHID Service")


udev = CTAP2HIDevice('/dev/uhid')
udev.start()

try:
    app = SysTrayApp() # runs until quit
finally: # Ensure clean shutdown
    logging.info("Waiting for UHID device thread to terminate...")
    udev.join(timeout=5)
    if udev.is_alive():
        logging.warning("UHID device thread did not terminate gracefully")
    else:
        logging.info("UHID device thread terminated successfully")



from setuptools import setup, Command
import os

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        for root, dirs, files in os.walk("./", topdown=False):
            for name in files:
                if name.endswith((".pyc", ".tgz", ".whl")):
                    print("remove {}".format(os.path.join(root, name)))
                    os.remove(os.path.join(root, name))
            for name in dirs:
                if name.endswith((".egg-info", "build", "dist", "__pycache__", "html")):
                    print("remove {}".format(os.path.join(root, name)))
                    #os.rmdir(os.path.join(root, name))
                    os.system('rm -vrf {}'.format(os.path.join(root, name)))


BUILD_ID = os.environ.get('TRAVIS_BUILD_NUMBER', 0)
if BUILD_ID == 0:
    BUILD_ID = os.environ.get('GITHUB_RUN_NUMBER', 0)

setup(
    name='soft_fido2',
    version='0.3.%s' % BUILD_ID,
    description='Software FIDO2 Passkey Authenticator',
    author='Lachlan Gleeson',
    author_email='lgleeson@au1.ibm.com',
    license='MIT',
    packages=["soft_fido2"],
    install_requires=[
        'cbor2>=4.1.2',
        'cryptography>=38.0.1',
        'asn1>=2.2.0',
        'PyJwt>=0.6.1'
        #'liboqs-python' #TODO use cryptography bindings for pqc
    ],
    url='https://github.com/lachlan-ibm/soft-fido2',
    project_urls={
        'Homepage': 'https://github.com/lachlan-ibm/soft-fido2',
        'Documentation': 'https://github.com/lachlan-ibm/soft-fido2/README.md',
        'Source': 'https://github.com/lachlan-ibm/soft-fido2',
        'Tracker': 'https://github.com/lachlan-ibm/soft-fido2/issues'
    },
    zip_safe=False,
        cmdclass={
        'clean': CleanCommand,
    },
    long_description=long_description,
    long_description_content_type='text/markdown'
)
